from six import print_
from six.moves import tkinter as tk
from six.moves import tkinter_tksimpledialog as tksimpledialog

class BrocoliDialog(tksimpledialog.Dialog, object):
    def buttonbox(self):
        # tweak to allow body expansion inside the simpledialog
        self.initial_focus.master.pack(padx=5, pady=5, expand=True,
                                       fill=tk.BOTH, side=tk.TOP)
        self.initial_focus.master.columnconfigure(1, weight=1)

        super(BrocoliDialog, self).buttonbox()



def option_is_true(option):
    option = option.lower()
    if option in ['1', 'yes', 'on', 'true']:
        return True
    elif option in ['0', 'no', 'off', 'false']:
        return False

    msg = 'invalid boolean option  value: {}'.format(option)
    raise ValueError(msg)


def option_is_false(option):
    return not option_is_true(option)




from . import catalog
from . progress_dialog import ProgressDialog
from . progress_dialog import progress_from_generator as progress
from . progress_dialog import unbounded_progress_from_generator as uprogress
from . exceptions import handle_catalog_exceptions
from . import exceptions
from . import navbar
from . listmanager import ColumnDef

import six
from six import print_

from six.moves import tkinter as tk
from six.moves import tkinter_tkfiledialog as filedialog
from six.moves import tkinter_tksimpledialog as simpledialog
from six.moves import tkinter_ttk as ttk
from six.moves import tkinter_messagebox as messagebox

import collections
import re


class TreeWidget(tk.Frame):
    """
    The main Brocoli widget displaying Catalog directory contents inside a
    ttk.TreeView
    """
    __placeholder_prefix = '__placeholder_'
    __empty_prefix = '__empty_'
    __dot_prefix = 'dot_'
    __dotdot_prefix = 'dotdot_'

    __prefix_path_re = re.compile('^(?P<prefix>{})(?P<suffix>.*)$'.format('|'.join([
        __placeholder_prefix,
        __empty_prefix,
        __dot_prefix,
        __dotdot_prefix,
    ])))
    __context_menu_upload = 'Upload local files'
    __context_menu_upload_directory = 'Recursive upload'
    __context_menu_download = 'Download to local disk'
    __context_menu_delete = 'Delete'
    __context_menu_mkdir = 'New directory'
    __context_menu_goto = 'Go to'
    __context_menu_properties = 'Properties'

    columns_def = collections.OrderedDict([
        ('#0', ColumnDef('#0', 'path')),
        ('user', ColumnDef('user', 'owner')),
        ('size', ColumnDef('size', 'size')),
        ('nreplicas', ColumnDef('nreplicas', '# of replicas')),
        ('mtime', ColumnDef('mtime', 'modification time')),
    ])

    def __init__(self, master):
        tk.Frame.__init__(self, master)

        self.master = master
        self.root_path = ''

        self.catalog = None
        self.path = None

        self.columns = [
            'user',
            'size',
            'nreplicas',
            'mtime',
        ]

        self.navigation_bar = navbar.NavigationBar(self, self.root_path,
                                                   self.set_path)
        self.navigation_bar.refresh_but.config(command=self.refresh)

        self.navigation_bar.grid(row=0, columnspan=2, sticky='ew')

        self.tree = ttk.Treeview(self, columns=self.columns)

        ysb = ttk.Scrollbar(self, orient='vertical', command=self.tree.yview)
        xsb = ttk.Scrollbar(self, orient='horizontal', command=self.tree.xview)
        self.tree.configure(yscroll=ysb.set, xscroll=xsb.set)

        for c in ['#0'] + self.columns:
            cd = self.columns_def[c]
            self.tree.heading(c, text=cd.text, anchor=cd.anchor)

        self.tree.bind('<<TreeviewOpen>>', self.open_cb)

        self.tree.grid(row=1, column=0, sticky='nsew')
        ysb.grid(row=1, column=1, sticky='ns')
        xsb.grid(row=2, column=0, sticky='ew')

        self.rowconfigure(1, weight=1)
        self.columnconfigure(0, weight=1)

        self._set_context_menu()

    def get_display_columns(self):
        return self.tree.config(cnf='displaycolumns')[-1]

    def set_display_columns(self, columns):
        if columns is None:
            columns = list(self.columns)

        self.tree.config(displaycolumns=columns)

    def set_connection(self, catalog_factory, path):
        try:
            # build catalog
            catalog = catalog_factory(self)

            if catalog is None:
                # user must have cancelled something
                return False

            path = catalog.normpath(path)

            # verify root path is valid
            if not catalog.isdir(path):
                messagebox.showerror('Path error',
                                     ('Path \'{}\' is not a ' +
                                      'directory').format(path))
                catalog.close()
                return False

        except IOError as e:
            if e.errno == exceptions.errno.ENOENT:
                messagebox.showerror('Connection error',
                                     ('Connection root path \'{}\' does ' +
                                      'not exist on catalog').format(path))
            else:
                messagebox.showerror('Connection error',
                                     ('Connection Error: {}').format(str(e)))
            return False
        except (exceptions.ConnectionError, exceptions.NetworkError) as e:
            messagebox.showerror('Connection error',
                                 ('Connection failed with error: ' +
                                  '{}').format(str(e)))

            return False

        self.catalog = catalog
        self.root_path = path

        self.set_path(path, clear_history=True)

        return True

    def set_path(self, path, clear_history=False):
        def is_subpath(path):
            if path == '':
                return False
            if path == self.root_path:
                return True
            return is_subpath(self.catalog.dirname(path))

        if path == self.path:
            return True, self.path

        if not is_subpath(path):
            # enforce path to be a subdirectory of connection root path
            return False, self.path

        try:
            if not self.catalog.isdir(path):
                messagebox.showerror('Path error',
                                     ('Path \'{}\' is not a ' +
                                      'directory').format(path))
                return False, self.path

        except IOError as e:
            if e.errno == exceptions.errno.ENOENT:
                messagebox.showerror('Connection error',
                                     ('Path \'{}\' does ' +
                                      'not exist on catalog').format(path))
                return False, self.path

            raise

        self.path = path

        self.navigation_bar.set_path(self.path, clear_history)

        self.refresh()

        return True, self.path

    def refresh(self):
        if self.catalog is None:
            return

        print_('refresh', self.path)

        for child in self.tree.get_children():
            self.tree.delete(child)

        self.process_directory('', self.path)

    def _set_context_menu(self):
        self.context_menu = tk.Menu(self.tree, tearoff=False)

        def _map(e):
            former_sel = self.get_selection()

            def selchanged(e):
                if former_sel != self.get_selection():
                    self.context_menu.unpost()

            self.context_menu_mapped = True
            self.tree.bind('<<TreeviewSelect>>', selchanged)

        def _unmap(e):
            self.context_menu_mapped = False
            self.tree.bind('<<TreeviewSelect>>')

        self.context_menu_mapped = False
        self.context_menu.bind('<Map>', _map)
        self.context_menu.bind('<Unmap>', _unmap)

        self.context_menu.add_command(label=self.__context_menu_mkdir,
                                      command=self.mkdir)
        self.context_menu.add_command(label=self.__context_menu_download,
                                      command=self.download)
        self.context_menu.add_command(label=self.__context_menu_upload,
                                      command=self.upload)
        self.context_menu.add_command(label=self.
                                      __context_menu_upload_directory,
                                      command=self.upload_directory)
        self.context_menu.add_command(label=self.__context_menu_delete,
                                      command=self.delete)

        self.context_menu.add_command(label=self.__context_menu_goto,
                                      command=self.goto_selected)

        self.context_menu.add_command(label=self.__context_menu_properties,
                                      command=self.selected_properties)

        self.tree.bind('<Double-Button-1>', self._goto, add='+')

        if (self.tree.tk.call('tk', 'windowingsystem') == 'aqua'):
            self.tree.bind('<Button-2>', self._select_and_pop)
            self.tree.bind('<Control-Button-1>', self._select_and_pop)
        else:
            self.tree.bind('<Button-3>', self._select_and_pop)

    def get_selection(self):
        ret = []

        for e in self.tree.selection():
            if e.startswith('__'):
                continue
            ret.append(e)

        return ret

    def _select_and_pop(self, e):
        item = self.tree.identify_row(e.y)
        if not item:
            return

        selection = self.get_selection()
        if item in selection:
            if self.context_menu_mapped:
                self.context_menu.unpost()
                return
        else:
            if selection and self.context_menu_mapped:
                self.context_menu.unpost()
                self.tree.selection_set(item)
                return
            self.tree.selection_set(item)

        selection = self.get_selection()
        state = tk.DISABLED
        is_directory = False
        if len(selection) == 1:
            item = selection[0]
            is_directory = (len(self.tree.get_children(item)) >= 1 or
                            item.startswith(self.__dot_prefix) or
                            item.startswith(self.__dotdot_prefix))

            if is_directory:
                state = tk.ACTIVE

        self.context_menu.entryconfig(self.__context_menu_upload, state=state)
        self.context_menu.entryconfig(self.__context_menu_upload_directory,
                                      state=state)
        self.context_menu.entryconfig(self.__context_menu_mkdir, state=state)

        self.context_menu.entryconfig(self.__context_menu_goto, state=state)

        state = tk.ACTIVE
        if (item.startswith(self.__empty_prefix) or
                item.startswith(self.__dotdot_prefix)):
            state = tk.DISABLED

        self.context_menu.entryconfig(self.__context_menu_download,
                                      state=state)

        state = tk.ACTIVE
        if (item.startswith(self.__dot_prefix) or
                item.startswith(self.__dotdot_prefix)):
            state = tk.DISABLED

        self.context_menu.entryconfig(self.__context_menu_delete,
                                      state=state)

        self.context_menu.post(e.x_root, e.y_root)

    def item_path(self, item):
        m = TreeWidget.__prefix_path_re.match(item)
        if m is None:
            return item

        return m.group('suffix')

    def _goto(self, e):
        item = self.tree.identify_row(e.y)
        if not item or not item.startswith(TreeWidget.__dotdot_prefix):
            return

        self.set_path(self.catalog.dirname(self.path))

    def _split_files_and_directories(self, selection):
        files = []
        directories = []
        for s in selection:
            s = self.item_path(s)
            if len(self.tree.get_children(s)) > 0:
                directories.append(s)
            else:
                files.append(s)

        return files, directories

    @handle_catalog_exceptions
    def download(self):
        selection = self.get_selection()
        destdir = filedialog.askdirectory()
        if not destdir:
            return

        print_('downloading', selection, 'to', destdir)

        files, directories = self._split_files_and_directories(selection)

        if files:
            with catalog.OperationStatusList(files) as osl:
                progress(self.master, 'download {} files'.format(len(files)),
                         self.catalog.download_files(files, destdir, osl))

                nok = osl.filter(catalog.OperationStatus.nok_statuses())
                if len(nok) > 0:
                    raise Exception('something has gone wrong: %s' % nok)

        if directories:
            with catalog.OperationStatusList() as osl:
                progress(self.master,
                         'download {} directories'.format(len(directories)),
                         self.catalog.download_directories(directories, destdir, osl))

                nok = osl.filter(catalog.OperationStatus.nok_statuses())
                if len(nok) > 0:
                    raise Exception('something has gone wrong: %s' % nok)

    @handle_catalog_exceptions
    def upload(self):
        path = self.item_path(self.get_selection()[0])
        files = filedialog.askopenfilenames()
        if not files:
            return

        print_('uploading', files, 'to', path)
        if files:
            with catalog.OperationStatusList(files) as osl:
                progress(self.master, 'upload {} files'.format(len(files)),
                         self.catalog.upload_files(files, path, osl))

                nok = osl.filter(catalog.OperationStatus.nok_statuses())
                if len(nok) > 0:
                    raise Exception('something has gone wrong: %s' % nok)

        pathid = path
        if path == self.path:
            pathid = ''
        self.process_directory(pathid, path)

    @handle_catalog_exceptions
    def upload_directory(self):
        path = self.item_path(self.get_selection()[0])
        directory = filedialog.askdirectory()
        if not directory:
            return

        print_('recursively uploading', directory, 'to', path)

        with catalog.OperationStatusList() as osl:
            progress(self.master, 'recursively upload {} directory'.format(1),
                     self.catalog.upload_directories((directory, ), path, osl))

            nok = osl.filter(catalog.OperationStatus.nok_statuses())
            if len(nok) > 0:
                raise Exception('something has gone wrong: %s' % nok)

        pathid = path
        if path == self.path:
            pathid = ''
        self.process_directory(pathid, path)

    @handle_catalog_exceptions
    def delete(self):
        selection = self.get_selection()

        print_('deleting', selection)

        files, directories = self._split_files_and_directories(selection)

        if len(selection) > 0:
            msg = 'Delete '
            msg_list = []
            if len(files) > 0:
                msg_list.append('{} files'.format(len(files)))
            if len(directories) > 0:
                msg_list.append('{} directories'.format(len(directories)))
            msg += ' and '.join(msg_list) + '?'
            if not messagebox.askokcancel('Confirm Delete', msg):
                return

        parents = {self.tree.parent(f) for f in selection}

        with ProgressDialog(self.master, '') as progress_bar, \
                catalog.OperationStatusList(files + directories) as osl:

            if files:
                n = len(files)
                progress_bar.set_message('delete {} files'.format(n))
                progress_bar.set(0, n)

                for p, n in self.catalog.delete_files(files, osl):
                    progress_bar.set(p, n)

                    if progress_bar.interrupted:
                        directories = []
                        break

            if directories:
                n = len(directories)
                progress_bar.set_message('delete {} directories'.format(n))
                progress_bar.set(0, n)

                for p, n in self.catalog.delete_directories(directories, osl):
                    progress_bar.set(p, n)

                    if progress_bar.interrupted:
                        break

        if '' in parents:
            self.process_directory('', self.path)
        else:
            for parent in parents:
                self.process_directory(parent, parent)

    @handle_catalog_exceptions
    def mkdir(self):
        selected = self.get_selection()[0]
        parent = self.item_path(selected)
        name = simpledialog.askstring('Create new directory',
                                      '{} directory name'.format(parent))
        if not name:
            return

        new_dir = self.catalog.join(parent, name)
        self.catalog.mkdir(new_dir)

        if selected.startswith(TreeWidget.__dot_prefix):
            selected = ''
        elif not self.tree.item(selected, option='open'):
            # non root closed directory parent needs to be open for display
            self.process_directory(selected, parent)
            self.tree.item(selected, open=True)
            return

        if self.__empty_prefix + parent in self.tree.get_children(selected):
            self.tree.delete(self.__empty_prefix + parent)

        st = self.catalog.lstat(new_dir)
        self.__fill_item(selected, parent, name, st)

    def goto_selected(self):
        selected = self.get_selection()[0]
        path = self.item_path(selected)
        print_('go to', selected, path)
        self.set_path(path)

    def selected_properties(self):
        selected = self.get_selection()[0]
        path = self.item_path(selected)

        props = None
        entry_type = ''
        if len(self.tree.get_children(selected)) > 0 or selected != path:
            # directories have children or have their iid different from
            # their path
            props = self.catalog.directory_properties(path)
            entry_type = 'Directory'
        else:
            # file properties
            props = self.catalog.file_properties(path)
            entry_type = 'File'

        if props is None or len(props) == 0:
            return

        tl = tk.Toplevel(self.master)
        tl.title(entry_type + ' properties: ' + path)

        nb = ttk.Notebook(tl)
        nb.pack(fill=tk.BOTH, expand=1)

        for t, p in props.items():
            nb.add(p.get_widget(nb), text=t, sticky='nsew')

        tl.wait_window()

    def open_cb(self, event):
        iid = self.tree.focus()
        children = self.tree.get_children(iid)
        if len(children) == 1 and \
           children[0].startswith(self.__placeholder_prefix):
            self.process_directory(iid, iid)

    def __fill_item(self, parent, path, name, st):
        abspath = self.catalog.join(path, name)

        values = [st[k] for k in self.columns]
        oid = self.tree.insert(parent, 'end', iid=abspath, text=name,
                               open=False, values=values)

        if st['isdir']:
            self.tree.insert(oid, 'end',
                             iid=self.__placeholder_prefix + abspath)

    def process_directory(self, parent, path):

        entries = self.catalog.listdir(path)

        item_children = self.tree.get_children(parent)
        for child in item_children:
            self.tree.delete(child)

        if parent == '':
            if path != self.root_path:
                ppath = self.catalog.dirname(path)
                self.tree.insert(parent, 'end',
                                 iid=self.__dotdot_prefix + ppath,
                                 text='..')
            self.tree.insert(parent, 'end', iid=self.__dot_prefix + path,
                             text='.')

        if not entries:
            self.tree.insert(parent, 'end', iid=self.__empty_prefix + path,
                             text='<empty dir>')
            return

        for k, v in entries.items():
            self.__fill_item(parent, path, k, v)



import os

import socket
import struct
import threading
from concurrent.futures import ThreadPoolExecutor

import M2Crypto

from irods.manager.data_object_manager import DataObjectManager

from irods.models import DataObject
import irods.exception as ex
import irods.keywords as kw
from irods.data_object import iRODSDataObject, irods_dirname, irods_basename
from irods.message import (
    iRODSMessage, FileOpenRequest, ObjCopyRequest, StringStringMap, DataObjInfo, ModDataObjMeta, Message, IntegerProperty, StringProperty, SubmessageProperty)


class ModifiedDataObjectManager(DataObjectManager):
    def get(self, path, file=None, **options):
        parent = self.sess.collections.get(irods_dirname(path))

        # TODO: optimize
        if file:
            self._download(path, file, **options)

        query = self.sess.query(DataObject)\
            .filter(DataObject.name == irods_basename(path))\
            .filter(DataObject.collection_id == parent.id)\
            .add_keyword(kw.ZONE_KW, path.split('/')[1])

        results = query.all() # get up to max_rows replicas
        if len(results) <= 0:
            raise ex.DataObjectDoesNotExist()
        # workaround incompatibility of irodsclient-0.8.2 and iRODS server 3.x
        for r in results:
            if DataObject.resc_hier not in r:
                r[DataObject.resc_hier] = None
        return iRODSDataObject(self, parent, results)


#define PortList_PI "int portNum; int cookie; int sock; int windowSize;
# str hostAddr[LONG_NAME_LEN];"


class PortList(Message):
    _name = 'PortList_PI'
    portNum = IntegerProperty()
    cookie = IntegerProperty()
    sock = IntegerProperty()
    windowSize = IntegerProperty()
    hostAddr = StringProperty()


# define PortalOprOut_PI "int status; int l1descInx; int numThreads;
# str chksum[NAME_LEN]; struct PortList_PI;"


class PortalOprResponse(Message):
    _name = 'PortalOprOut_PI'
    status = IntegerProperty()
    l1descInx = IntegerProperty()
    numThreads = IntegerProperty()
    chksum = StringProperty()
    PortList_PI = SubmessageProperty(PortList)

# define INT_PI "int myInt;"


class OprComplete(Message):
    _name = 'INT_PI'
    myInt = IntegerProperty()


import irods.exception as ex
from irods.api_number import api_number
from irods import DEFAULT_CONNECTION_TIMEOUT
import irods.keywords as kw

class LockCounter:
    def __init__(self, value=0):
        self.lock = threading.Lock()
        self.__value = value

    def __get_value(self):
        return self.__value

    def __set_value(self, new_value):
        with self.lock:
            self.__value = new_value
            return self.__value

    value = property(__get_value, __set_value)

    def __add(self, add):
        with self.lock:
            self.__value += add
            return self.__value

    def incr(self, increment=1):
        return self.__add(increment)

    def decr(self, decrement=1):
        return self.__add(- decrement)

def connect_to_portal(host, port, cookie,
                      timeout=DEFAULT_CONNECTION_TIMEOUT):
    address = (host, port)
    try:
        s = socket.create_connection(address, timeout)
    except socket.error:
        raise ex.NetworkException(
            "Could not connect to specified host and port: " +
            "{}:{}".format(*address))

    fmt = '!i'
    sent = s.send(struct.pack(fmt, cookie))

    if sent != struct.calcsize(fmt):
        s.close()
        raise ex.NetworkException(
            "SYS_PORT_COOKIE_ERR: {}:{}".format(*address))

    return s

def recv_xfer_header(sock):
    # typedef struct TransferHeader { int oprType; int flags;
    # rodsLong_t offset; rodsLong_t length; } transferHeader_t;

    fmt = '!iiqq'
    size = struct.calcsize(fmt)
    buf = bytearray(size)
    recv_size = sock.recv_into(buf, size)
    if recv_size != size:
        try:
            # try to unpack only opr type instead of whole header
            opr = struct.unpack('!i', buf[0:recv_size])[0]
            return (opr, 0, 0, 0)
        except:
            # raise error if unpack failed
            raise ex.SYS_COPY_LEN_ERR

    u = struct.unpack(fmt, buf)
    return u

class Encryption:
    def __init__(self, connection):
        self.algorithm = connection.account.encryption_algorithm.lower().replace('-', '_')
        self.key = connection.shared_secret
        self.key_size = connection.account.encryption_key_size

        self.ifmt = 'i'
        self.isize = struct.calcsize(self.ifmt)
        self.ibuf = bytearray(self.isize)

    def recv_int(self, sock):
        recv_size = sock.recv_into(self.ibuf, self.isize)

        if recv_size != self.isize:
            raise ex.SYS_COPY_LEN_ERR

        u = struct.unpack(self.ifmt, self.ibuf)
        return u[0]

    def send_int(self, sock, i):
        struct.pack_into(self.ifmt, self.ibuf, 0, i)

        sock.sendall(self.ibuf)

    def generate_key(self):
        return os.urandom(self.key_size)

    def __xxcrypt(self, iv, buf, op):
        cipher = M2Crypto.EVP.Cipher(alg=self.algorithm, key=self.key,
                                     iv=iv, op=op)
        return cipher.update(buf) + cipher.final()

    def decrypt(self, buf):
        iv = buf[0:self.key_size]
        text = buf[self.key_size:]
        try:
            return self.__xxcrypt(iv, text, op=0)
        except TypeError:
            # Python 2 doesn't seem to know that memoryview on bytearray
            # are bytelike objects...
            return self.__xxcrypt(bytearray(iv), bytearray(text), op=0)

    def encrypt(self, iv, buf):
        return iv + self.__xxcrypt(iv, buf, op=1)


class MultithreadDataObjectManager(ModifiedDataObjectManager):

    # lib/api/include/dataObjInpOut.h
    DONE_OPR = 9999

    def download_parallel(self, irods_path, local_path, executor=None,
                          progress_cb=None, **options):

        progress_cb = progress_cb or (lambda l, i, c: True)

        def opr_complete(conn, desc):
            message = iRODSMessage('RODS_API_REQ',
                                   OprComplete(myInt=desc),
                                   int_info=api_number['OPR_COMPLETE_AN'])

            conn.send(message)
            return conn.recv()

        def recv_task(sock, local_path, conn, task_count):
            try:
                with open(local_path, 'r+b') as lf:
                    buf = memoryview(bytearray(self.READ_BUFFER_SIZE))
                    if use_encryption:
                        crypt = Encryption(conn)

                    while True:
                        opr, _, offset, size = recv_xfer_header(sock)
                        if opr == self.DONE_OPR:
                            break

                        lf.seek(offset)

                        while size > 0:
                            if task_count.value < 0:
                                return

                            to_read = min(size, self.READ_BUFFER_SIZE)

                            if use_encryption:
                                to_read = crypt.recv_int(sock)

                            all_read = 0
                            while all_read < to_read:
                                current = buf[all_read:]
                                read_size = sock.recv_into(current,
                                                           to_read - all_read)
                                all_read += read_size

                            plaintext = buf[0:all_read]

                            if use_encryption:
                                plaintext = crypt.decrypt(plaintext)
                                all_read = len(plaintext)

                            lf.write(plaintext)
                            size -= all_read

                            if not progress_cb(local_path, irods_path,
                                               all_read):
                                task_count.value = -1
                                conn.release()
            finally:
                sock.close()

            if task_count.decr() == 0:
                # last task has to complete iRODS operation
                opr_complete(conn, desc)
                conn.release()

        def write_from_response(irods_path, local_path, response, conn,
                                progress_cb):
            try:
                with open(local_path, 'wb') as lf:
                    lf.write(response.bs)

                opr_complete(conn, desc)
            finally:
                conn.release()
                progress_cb(local_path, irods_path, len(response.bs))

            return []

        # Check for force flag if local file exists
        if os.path.exists(local_path) and kw.FORCE_FLAG_KW not in options:
            raise ex.OVERWRITE_WITHOUT_FORCE_FLAG

        response, message, conn = self._open_request(irods_path,
                                                     'DATA_OBJ_GET_AN',
                                                     'r', 0, **options)

        use_encryption = hasattr(conn, 'shared_secret') and conn.shared_secret is not None

        desc = message.l1descInx

        if desc <= 2:
            # file contents are directly embeded in catalog response

            if executor is not None:
                return [executor.submit(write_from_response, irods_path,
                                        local_path, response, conn,
                                        progress_cb)]

            write_from_response(irods_path, local_path, response, conn,
                                progress_cb)
            return []

        futs = []

        nt = message.numThreads
        if nt <= 0:
            nt = 1

        host = message.PortList_PI.hostAddr
        port = message.PortList_PI.portNum
        cookie = message.PortList_PI.cookie

        task_count = LockCounter(nt)

        join = False
        if executor is None:
            # handle parallel transfer with own executor
            executor = ThreadPoolExecutor(max_workers=nt)
            join = True

        with open(local_path, 'w'):
            # create local file
            pass

        for _ in range(nt):
            sock = connect_to_portal(host, port, cookie)
            fut = executor.submit(recv_task, sock, local_path, conn,
                                  task_count)

            futs.append(fut)

        if join:
            executor.shutdown()
            exceptions = []
            for f in futs:
                e=f.exception()
                if e is not None:
                    exceptions.append(e)
            if len(exceptions) > 0:
                msgs = ['%s%s' % (type(e).__name__, str(e)) for e in exceptions]
                raise Exception(', '.join(msgs))

        return futs

    def _put_opened_file(self, local_path, obj):
        with open(local_path, 'rb') as f:
            for chunk in chunks(f, self.WRITE_BUFFER_SIZE):
                obj.write(chunk)

    def put_parallel(self, local_path, irods_path, executor=None,
                     progress_cb=None, **options):

        progress_cb = progress_cb or (lambda l, i, c: True)

        def send_task(sock, local_path, conn, task_count):
            try:
                with open(local_path, 'rb') as lf:
                    if use_encryption:
                        crypt = Encryption(conn)

                    while True:
                        opr, _, offset, size = recv_xfer_header(sock)

                        if opr == self.DONE_OPR:
                            break

                        lf.seek(offset)

                        if use_encryption:
                            iv = crypt.generate_key()

                        while size > 0:
                            if task_count.value < 0:
                                return
                            to_read = min(size, self.WRITE_BUFFER_SIZE)

                            buf = lf.read(to_read)
                            read_size = len(buf)

                            new_size = read_size
                            if use_encryption:
                                buf = crypt.encrypt(iv, buf)
                                new_size = len(buf)
                                crypt.send_int(sock, new_size)

                            sock.sendall(buf)

                            size -= read_size

                            if not progress_cb(local_path, irods_path,
                                               read_size):
                                task_count.value = -1
            finally:
                sock.close()

            if task_count.decr() == 0:
                # last task has to complete iRODS operation
                message = iRODSMessage('RODS_API_REQ',
                                       OprComplete(myInt=desc),
                                       int_info=api_number['OPR_COMPLETE_AN'])

                conn.send(message)
                conn.recv()
                conn.release()

                replicate()

        def send_task_cb(fut):
            if fut.exception() is not None:
                # exception occurred in send_task. Mark other tasks for exit
                task_count.value = -1

        def send_to_catalog(conn, local_path, irods_path, desc, **options):
            with io.BufferedRandom(iRODSDataObjectFileRaw(conn,
                                   desc, **options)) as o:
                self._put_opened_file(local_path, o)
            replicate()

        def replicate():
            if kw.ALL_KW in options:
                options[kw.UPDATE_REPL_KW] = ''
                self.replicate(irods_path, **options)

        local_size = os.lstat(local_path).st_size

        # Set operation type to trigger acPostProcForPut
        if kw.OPR_TYPE_KW not in options:
            options[kw.OPR_TYPE_KW] = 1 # PUT_OPR

        _, message, conn = self._open_request(irods_path,
                                              'DATA_OBJ_PUT_AN', 'w',
                                              local_size, **options)

        use_encryption = conn.shared_secret is not None

        desc = message.l1descInx
        nt = message.numThreads
        if nt <= 0:
            nt = 1

        futs = []
        join = False

        if executor is None:
            if nt <= 1:
                # sequential put
                send_to_catalog(conn, local_path, irods_path, desc,
                                **options)
                return []

            # handle parallel transfer with own executor
            executor = ThreadPoolExecutor(max_workers=nt)
            join = True

        if nt <= 1:
            fut = executor.submit(send_to_catalog, conn, local_path,
                                  irods_path, desc, **options)
            futs.append(fut)
        else:
            host = message.PortList_PI.hostAddr
            port = message.PortList_PI.portNum
            cookie = message.PortList_PI.cookie

            task_count = LockCounter(nt)

            for _ in range(nt):
                sock = connect_to_portal(host, port, cookie)
                fut = executor.submit(send_task, sock, local_path, conn,
                                      task_count)
                fut.add_done_callback(send_task_cb)

                futs.append(fut)

        if join:
            executor.shutdown()
            exceptions = []
            for f in futs:
                e=f.exception()
                if e is not None:
                    exceptions.append(e)
            if len(exceptions) > 0:
                msgs = ['%s%s' % (type(e).__name__, str(e)) for e in exceptions]
                raise Exception(', '.join(msgs))

        return futs

    def _open_request(self, path, an_name, mode, size, **options):
        if kw.DEST_RESC_NAME_KW not in options:
            # Use client-side default resource if available
            try:
                options[kw.DEST_RESC_NAME_KW] = self.sess.default_resource
            except AttributeError:
                pass

        try:
            oprType = options[kw.OPR_TYPE_KW]
        except KeyError:
            oprType = 0

        flags, seek_to_end = {
            'r': (self.O_RDONLY, False),
            'r+': (self.O_RDWR, False),
            'w': (self.O_WRONLY | self.O_CREAT | self.O_TRUNC, False),
            'w+': (self.O_RDWR | self.O_CREAT | self.O_TRUNC, False),
            'a': (self.O_WRONLY | self.O_CREAT, True),
            'a+': (self.O_RDWR | self.O_CREAT, True),
        }[mode]
        # TODO: Use seek_to_end

        message_body = FileOpenRequest(
            objPath=path,
            createMode=0,
            openFlags=flags,
            offset=0,
            dataSize=size,
            numThreads=self.sess.numThreads,
            oprType=oprType,
            KeyValPair_PI=StringStringMap(options),
        )

        message = iRODSMessage('RODS_API_REQ', msg=message_body,
                               int_info=api_number[an_name])

        conn = self.sess.pool.get_connection()
        conn.send(message)
        resp = conn.recv()

        return resp, resp.get_main_message(PortalOprResponse), conn




import os
import shutil
from datetime import datetime
from collections import OrderedDict

from six import print_


class Catalog(object):
    """
    Base class for catalog objects. All methods have to be overridden by
    sub-classes.
    """
    def lstat(self, path):
        """
        Returns a dictionary of informations about specified path. Mandatory
        fields are: size, mtime, nreplicas, user, isdir
        """
        raise NotImplementedError

    def listdir(self, path):
        """
        Returns directory contents (only filenames, not absolute paths) in the
        form of a dictionary. Each value of the dictionary has to be itself a
        dictionary as would be returned by lstat()
        """
        raise NotImplementedError

    def isdir(self, path):
        """
        Returns wether the specified path is a directory.
        """
        raise NotImplementedError

    def join(self, *args):
        """
        Performs the equivalent of os.path.join() in the context of the
        Catalog object.
        """
        raise NotImplementedError

    def splitname(self, path):
        """
        Performs the equivalent of os.path.split() in the context of the
        Catalog object.
        """
        raise NotImplementedError

    def dirname(self, path):
        """
        Performs the equivalent of os.path.dirname() in the context of the
        Catalog object.
        """
        raise NotImplementedError

    def basename(self, path):
        """
        Performs the equivalent of os.path.basename() in the context of the
        Catalog object.
        """
        raise NotImplementedError

    def normpath(self, path):
        """
        Performs the equivalent of os.path.normpath() in the context of the
        Catalog object.
        """
        raise NotImplementedError


    def download_files(self, pathlist, destdir, osl):
        """
        Downloads files from the catalog pathlist to local destdir.
        """
        raise NotImplementedError

    def download_directories(self, pathlist, destdir, osl):
        """
        Downloads directories contents to local destdir.
        """
        raise NotImplementedError

    def upload_files(self, files, path, osl):
        """
        Uploads local files to catalog destination path (a directory).
        """
        raise NotImplementedError

    def upload_directories(self, dirs, path, osl):
        """
        Uploads local directories content to catalog destination path (a
        directory).
        """
        raise NotImplementedError

    def delete_files(self, files, osl):
        """
        Deletes catalog files.
        """
        raise NotImplementedError

    def delete_directories(self, directories, osl):
        """
        Recusively deletes catalog directories.
        """
        raise NotImplementedError

    def mkdir(self, path):
        """
        Creates a new catalog directory.
        """
        raise NotImplementedError

    def directory_properties(self, path):
        """
        Returns a dictionary of properties for a directory path
        """
        raise NotImplementedError

    def file_properties(self, path):
        """
        Returns a dictionary of properties for a file path
        """
        raise NotImplementedError

    def close(self):
        pass

    @classmethod
    def config_fields(cls):
        """
        Returns a dictionary of form.FormField objects for catalog object
        configuration.
        """
        raise NotImplementedError


class OperationStatus(object):
    NEW = 0
    IN_PROGRESS = 1
    DONE = 2
    FAILED = 3
    INTERRUPTED = 4
    reverse_status = {v: k for k, v in locals().items() if not k.startswith('__')}

    def __init__(self, size=0, cancel=None):
        self.status = self.NEW
        self.progress = 0
        self.size = size # work size
        self.file_size = size # actual file size
        self.exception = None

        self.cancel = cancel or (lambda e: None)

        self.current_element = None

    def in_progress(self, current_element):
        self.status = self.IN_PROGRESS
        self.current_element = current_element

    def done(self):
        self.status = self.DONE
        self.current_element = None

    def interrupt(self):
        if self.status not in (self.NEW, self.IN_PROGRESS):
            return

        self.status = self.INTERRUPTED
        if self.status != self.NEW:
            self.cancel(self)
        self.current_element = None

    def fail(self, exception=None):
        self.exception = exception
        self.status = self.FAILED
        self.cancel(self)
        self.current_element = None

    def print_status(self):
        return self.reverse_status[self.status]

    def __str__(self):
        ret = 'status=%s file_size=%s' % (self.print_status(), self.file_size)

        if self.status in (self.IN_PROGRESS, self.INTERRUPTED, self.FAILED):
            ret += ' progress=%s current_element=%s' % (self.progress, self.current_element)
            if self.status == self.FAILED:
                ret += ' exception=\'%s\'' % self.exception

        return '<' + ret + '>'

    @classmethod
    def statuses(cls):
        return {v for v in cls.reverse_status}

    @classmethod
    def ok_statuses(cls):
        return {cls.DONE}

    @classmethod
    def nok_statuses(cls):
        return cls.statuses() - cls.ok_statuses()

class OperationStatusList(OrderedDict):
    def __init__(self, keylist=[], cancel=None):
        items = [(k, OperationStatus(cancel=cancel)) for k in keylist]
        super(OperationStatusList, self).__init__(items)

    def finalize(self):
        for os in self.values():
            os.interrupt()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.finalize()

    def add_keys(self, keylist, cancel=None):
        items = [(k, OperationStatus(cancel=cancel)) for k in keylist]
        self.update(items)

    def update_list(self, keys, **kwargs):
        for attr, arg in kwargs.items():
            if isinstance(arg, dict):
                for k in keys:
                    setattr(self[k], attr, arg[k])
            else:
                for k in keys:
                    setattr(self[k], attr, arg)

    def size(self):
        return sum([o.size for o in self.values()])

    def filter(self, statuses):
        ret = OperationStatusList()

        for k, v in self.items():
            if v.status in statuses:
                ret[k] = v

        return ret

    def __str__(self):
        ret = ''

        l = ['\'%s\' : %s' % (k, v) for k, v in self.items()]

        return '{' + ', '.join(l) + '}'

class OSCatalog(Catalog):
    """
    Presents contents from local filesystem. Useful for debugging.
    """
    def lstat(self, path):
        stats = os.lstat(path)

        nreplicas = '1'
        isdir = False
        if self.isdir(path):
            nreplicas = ''
            isdir = True
        ret = {
            'user': stats.st_uid,
            'size': stats.st_size,
            'mtime': datetime.fromtimestamp(stats.st_mtime),
            'nreplicas': nreplicas,
            'isdir': isdir,
        }

        return ret

    def listdir(self, path):
        entries = os.listdir(path)

        directories = []
        files = []

        ret = {}
        for e in entries:
            v = self.lstat(self.join(path, e))
            ret[e] = v

        return ret

    def isdir(self, path):
        return os.path.isdir(path)

    def join(self, *args):
        return os.path.join(*args)

    def splitname(self, path):
        return os.path.split(path)

    def dirname(self, path):
        return os.path.dirname(path)

    def basename(self, path):
        return os.path.basename(path)

    def normpath(self, path):
        return os.path.normpath(path)

    def download_files(self, pathlist, destdir, osl):
        number = len(pathlist)
        for path in pathlist:
            osl[path].size = 1
            osl[path].cancel = os.unlink

        i = 0
        for path in pathlist:
            osl[path].in_progress(os.path.join(destdir, self.basename(path)))
            shutil.copy2(path, destdir)
            osl[path].done()
            i += 1
            yield i, number

    def download_directories(self, pathlist, destdir, osl):
        number = len(pathlist)
        for p in pathlist:
            osl[p].size = 1

        i = 0
        for path in pathlist:
            osl[path].in_progress(None)

            # shutil.copytree needs a fresh new destination directory
            ddir = os.path.join(destdir, os.path.basename(path))
            if os.path.exists(ddir):
                shutil.rmtree(ddir)
            shutil.copytree(path, ddir)

            osl[path].done()

            i += 1
            yield i, number

    def upload_files(self, files, path, osl):
        number = len(files)
        for f in files:
            osl[f].size = 1

        i = 0
        for f in files:
            osl[f].in_progress(None)
            shutil.copy2(f, path)
            osl[f].done()
            i += 1
            yield i, number

    def upload_directories(self, dirs, path, osl):
        number = len(dirs)
        for d in dirs:
            osl[d].size = 1

        i = 0
        for d in dirs:
            osl[d].in_progress(None)

            # shutil.copytree needs a fresh new destination directory
            ddir = os.path.join(path, os.path.basename(d))
            if os.path.exists(ddir):
                shutil.rmtree(ddir)

            shutil.copytree(d, ddir)

            osl[d].done()

            i += 1
            yield i, number

    def delete_files(self, files, osl):
        number = len(files)
        for f in files:
            osl[f].size = 1

        i = 0
        for f in files:
            osl[f].in_progress(None)
            os.unlink(f)
            osl[f].done()
            i += 1
            yield i, number


    def delete_directories(self, directories, osl):
        number = len(directories)
        for d in directories:
            osl[d].size = 1

        i = 0
        for d in directories:
            osl[d].in_progress(None)

            shutil.rmtree(d)

            osl[d].done()

            i += 1
            yield i, number

    def mkdir(self, path):
        os.mkdir(path)

    def directory_properties(self, path):
        return {}

    def file_properties(self, path):
        return {}

    @classmethod
    def config_fields(cls):
        return {}





from six import print_
from six.moves import tkinter as tk
from six.moves import tkinter_ttk as ttk

import collections
import copy

from . import form
from . dialog import BrocoliDialog
from . exceptions import handle_catalog_exceptions


class ColumnDef(object):
    def __init__(self, name, text, anchor='w', form_field=None, display=True):
        self.name = name
        self.text = text
        self.anchor = anchor
        self.form_field = form_field
        self.display = display


class ItemConfigDialog(BrocoliDialog):
    def __init__(self, master, item_fields, item_config=None, **kwargs):
        self.item_fields = item_fields
        self.item_config = item_config

        for ff in self.item_fields.values():
            ff.reset()

        super(ItemConfigDialog, self).__init__(master, **kwargs)

    def body(self, master):
        if self.item_config is not None:
            for k, ff in self.item_fields.items():
                if k in self.item_config:
                    ff.from_string(self.item_config[k])
                    ff.contents_from_config(self.item_config)

        self.item_frame = form.FormFrame(master)
        self.item_frame.grid_fields(self.item_fields.values(), False)
        self.item_frame.grid(row=0, column=1, sticky='nsew')
        self.item_frame.columnconfigure(1, weight=1)

        self.result = None

        return self.item_frame

    def apply(self):
        self.result = collections.OrderedDict()

        for k, ff in self.item_fields.items():
            self.result[k] = ff.to_string()
            self.result.update(ff.get_contents())


class ListManager(tk.Frame):
    def __init__(self, master, columns_def, add_cb=None, remove_cb=None,
                 edit_cb=None):
        tk.Frame.__init__(self, master)

        self.master = master
        self.columns_def = columns_def

        self.fields = collections.OrderedDict([
            (k, cd.form_field) for k, cd in self.columns_def.items()
        ])

        self.items = None

        column_ids = [
            k for k, cd in self.columns_def.items() if cd.display and k != '#0'
        ]

        self.tree = ttk.Treeview(self, columns=column_ids, selectmode='browse')

        ysb = ttk.Scrollbar(self, orient='vertical', command=self.tree.yview)
        xsb = ttk.Scrollbar(self, orient='horizontal', command=self.tree.xview)
        self.tree.configure(yscroll=ysb.set, xscroll=xsb.set)

        for c in ['#0'] + column_ids:
            cd = self.columns_def[c]
            self.tree.heading(c, text=cd.text, anchor=cd.anchor)

        self.tree.grid(row=0, column=0, sticky='nsew')
        ysb.grid(row=0, column=1, sticky='ns')
        xsb.grid(row=1, column=0, sticky='ew')

        butbox = tk.Frame(self)
        butbox.grid(row=0, column=2, sticky='ns')

        self.newbut = None
        self.removebut = None
        self.editbut = None
        self.duplicatebut = None

        self.add_cb = add_cb
        if add_cb is not None:
            self.newbut = tk.Button(butbox, text='Add', command=self.add)
            self.newbut.grid(row=0, column=0, sticky='ew')

        self.remove_cb = remove_cb
        if remove_cb is not None:
            self.removebut = tk.Button(butbox, text='Remove',
                                       command=self.remove, state=tk.DISABLED)
            self.removebut.grid(row=1, column=0, sticky='ew')

        self.edit_cb = edit_cb
        if edit_cb is not None:
            self.editbut = tk.Button(butbox, text='Edit', command=self.edit,
                                     state=tk.DISABLED)
            self.editbut.grid(row=2, column=0, sticky='ew')

            self.duplicatebut = tk.Button(butbox, text='duplicate',
                                          command=self.duplicate,
                                          state=tk.DISABLED)
            self.duplicatebut.grid(row=3, column=0, sticky='ew')

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        self.grid(sticky='nsew')

        self.tree.bind('<<TreeviewSelect>>', self.selchanged)

    def populate(self, rows):
        for child in self.tree.get_children():
            self.tree.delete(child)

        self.items = collections.OrderedDict([(v.get('iid', v['#0']), v)
                                              for v in rows])

        for v in rows:
            values = [v[c] for c in self.columns_def
                      if c != '#0' and c != 'iid']
            iid = v.get('iid', v['#0'])

            root_node = self.tree.insert('', 'end', iid=iid, text=v['#0'],
                                         open=True, values=values)

    @handle_catalog_exceptions
    def add(self):
        n = ItemConfigDialog(self.master, self.fields)
        new = n.result
        if new is None:
            return

        # add_cb should return new iid or None
        iid = self.add_cb(new)

        name = new['#0']
        iid = iid or name

        self.items[iid] = new
        values = [v for k, v in self.items[iid].items() if k != '#0']
        self.tree.insert('', 'end', iid=iid or name, text=name, values=values)

    def remove(self):
        selected = self.tree.selection()[0]

        self.remove_cb(self.items[selected])

        del self.items[selected]

        self.tree.selection_set('')
        self.tree.delete(selected)

    def edit(self, duplicate=False):
        selected = self.tree.selection()[0]
        item = self.items[selected]

        if duplicate:
            item = copy.deepcopy(item)
            item['#0'] = item['#0'] + ' (copy)'

        icd = ItemConfigDialog(self.master, self.fields, item)

        if icd.result is None:
            return

        new_name = icd.result['#0']

        # update list display
        self.items[new_name] = icd.result
        values = [v for k, v in icd.result.items() if k != '#0']

        duped = False

        if new_name != selected:
            if not duplicate:
                self.tree.delete(selected)
            else:
                duped = True

            self.tree.insert('', 'end', iid=new_name, text=new_name,
                             values=values)
        else:
            self.tree.item(selected, text=selected, values=values)

        if duped:
            self.add_cb(icd.result)
        else:
            self.edit_cb(icd.result)

    def duplicate(self):
        self.edit(duplicate=True)

    def selchanged(self, event):
        buts = [b for b in [self.editbut, self.removebut, self.duplicatebut]
                if b is not None]
        if self.tree.selection():
            for b in buts:
                b.config(state=tk.NORMAL)
        else:
            for b in buts:
                b.config(state=tk.DISABLED)


class List(object):
    def __init__(self, column_defs, rows=[], add_cb=None, remove_cb=None,
                 edit_cb=None):
        self.column_defs = column_defs
        self.rows = rows
        self.add_cb = add_cb
        self.remove_cb = remove_cb
        self.edit_cb = edit_cb

    def get_widget(self, master):
        lm = ListManager(master, self.column_defs, add_cb=self.add_cb,
                         remove_cb=self.remove_cb, edit_cb=self.edit_cb)
        lm.populate(self.rows)

        return lm



import os
import errno
import traceback

from six.moves import tkinter_messagebox as messagebox
from six import print_


class BrocoliError(Exception):
    def __init__(self, exception):
        self.exception = exception

    def __str__(self):
        return type(self.exception).__name__ + ': ' + str(self.exception)


class ConnectionError(BrocoliError):
    pass


class NetworkError(BrocoliError):
    pass


class FileNotFoundError(BrocoliError):
    pass


class CatalogLogicError(BrocoliError):
    pass

class ChecksumError(Exception):
    pass

def ioerror(no):
    return IOError(no, os.strerror(no))


def handle_catalog_exceptions(method):
    """
    Method decorator that presents Brocoli exceptions to the user with messages
    """
    def method_wrapper(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except ConnectionError as e:
            messagebox.showerror('Catalog Connection Error',
                                 ('Connection failed: ' +
                                  '{}').format(str(e)))
        except FileNotFoundError as e:
            messagebox.showerror('File Not Found',
                                 ('Catalog file was not found: ' +
                                  '{}').format(str(e)))
        except CatalogLogicError as e:
            messagebox.showerror('Catalog Logic Error',
                                 ('Catalog logic error occurred: ' +
                                  '{}').format(str(e)))
        except ChecksumError as e:
            messagebox.showerror('Checksum Error',
                                 ('Checksum error occurred: ' +
                                  '{}').format(str(e)))
        except Exception as e:
            messagebox.showerror('Unknown Error',
                                 ('Some unknown exception occurred: ' +
                                  '{}').format(str(e)))
            print_(traceback.format_exc())

    return method_wrapper



"""
Implements iRODS Catalog object and methods
"""

from . import catalog
from . import form
from . import exceptions
from . listmanager import ColumnDef, List
from . config_option import option_is_true

from . irodsdom import ModifiedDataObjectManager

import re
import os
import io
import hashlib
import base64
import collections
import datetime
import ssl
import time

from concurrent.futures import ThreadPoolExecutor, wait, as_completed, FIRST_COMPLETED
from queue import Queue
from threading import Lock

from six import print_
from six.moves import tkinter as tk

import irods
from irods.session import iRODSSession
from irods import password_obfuscation
from irods.manager.data_object_manager import DataObjectManager
from . irodsdom import MultithreadDataObjectManager
from irods.manager.collection_manager import CollectionManager
from irods.models import DataObject, Collection
from irods.manager import data_object_manager
from irods.data_object import chunks
from irods.column import Like
from irods.api_number import api_number
import irods.keywords as kw
import irods.constants as const
import irods.message as message
import irods.exception

_getuid = None
if hasattr(os, 'getuid'):
    _getuid = os.getuid
else:
    import getpass

    # generate a fake uid on systems that lacks os.getuid()
    def _fake_getuid():
        return int('0x' + hashlib.md5(getpass.getuser().encode()).hexdigest(),
                   16) % 10000

    _getuid = _fake_getuid


def parse_env3(path):
    """
    parse iRODS v3 iCommands environment files
    """

    pat = ("^\s*(?P<name>\w+)" + "(\s+|\s*=\s*)[\'\"]?(?P<value1>[^\'\"\n]*)" +
           "[\'\"]?\s*$")
    envre = re.compile(pat)

    ret = {}

    with open(path, 'r') as f:
        for l in f.readlines():
            m = envre.match(l)
            if m:
                ret[m.group("name")] = m.group("value1")

    return ret


def local_trees_stats(dirs):
    """
    Gathers stats (number of files and cumulated size) of sub-trees on a local
    directories list
    """
    total_nfiles = 0
    total_size = 0
    stats = {}

    for d in dirs:
        nfiles = 0
        size = 0

        for root, dirs, files in os.walk(d):
            nfiles += len(files)
            size += sum(os.path.getsize(os.path.join(root, name))
                        for name in files)
        stats[d] = size

        total_nfiles += nfiles
        total_size += size

    return total_nfiles, total_size, stats


def local_files_stats(files):
    """
    Computes stats (number of files and cumulated size) on a file list
    """
    stats = {f: os.path.getsize(f) for f in files}
    return len(files), sum(v for v in stats.values()), stats


def method_translate_exceptions(method):
    """
    Method decorator that translates iRODS to Brocoli exceptions
    """
    def method_wrapper(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except irods.exception.CAT_INVALID_AUTHENTICATION as e:
            self.close()
            raise exceptions.ConnectionError(e)
        except (irods.exception.NetworkException, ssl.SSLError) as e:
            raise exceptions.NetworkError(e)
        except irods.exception.CAT_UNKNOWN_COLLECTION as e:
            raise exceptions.FileNotFoundError(e)
        except irods.exception.CAT_SQL_ERR as e:
            raise exceptions.CatalogLogicError(e)
        except irods.exception.SYS_FILE_DESC_OUT_OF_RANGE as e:
            raise exceptions.CatalogLogicError(e)

    return method_wrapper


def function_translate_exceptions(func):
    """
    Function decorator that translates iRODS to Brocoli exceptions
    """
    def function_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except irods.exception.CAT_INVALID_AUTHENTICATION as e:
            raise exceptions.ConnectionError(e)
        except irods.exception.NetworkException as e:
            raise exceptions.NetworkError(e)
        except irods.exception.CAT_UNKNOWN_COLLECTION as e:
            raise exceptions.FileNotFoundError(e)
        except irods.exception.CAT_SQL_ERR as e:
            raise exceptions.CatalogLogicError(e)
        except irods.exception.SYS_FILE_DESC_OUT_OF_RANGE as e:
            raise exceptions.CatalogLogicError(e)

    return function_wrapper


def splitsize(size, n, bs):
    # generate a sequence of offsets and chunk size based on n (nb of threads)
    # and bs (block size)

    # number of blocks
    nb = max(0, (size - 1) // bs) + 1

    # ensure each thread has at least 2 chunks
    n = max(1, min(n, nb//2))

    d, m = divmod(nb, n)

    p = 0
    for i in range(m):
        yield p*bs, (d + 1)*bs
        p += d + 1

    for i in range(m, n):
        yield p*bs, min(d*bs, size - p*bs)
        p += d


class LockedSession:
    class LockedDOM:
        def __init__(self, dom, lock=None):
            self.dom = dom
            self.lock = lock or Lock()

        def open(self, *args, **kwargs):
            with self.lock:
                return self.dom.open(*args, **kwargs)

        def replicate(self, *args, **kwargs):
            with self.lock:
                return self.dom.replicate(*args, **kwargs)

        def truncate(self, *args, **kwargs):
            with self.lock:
                return self.dom.truncate(*args, **kwargs)

        def unlink(self, *args, **kwargs):
            with self.lock:
                return self.dom.unlink(*args, **kwargs)

        def get(self, *args, **kwargs):
            with self.lock:
                return self.dom.get(*args, **kwargs)

    def __init__(self, session, lock=None):
        self.session = session
        self.lock = lock or Lock()

    @property
    def data_objects(self):
        return self.LockedDOM(MultithreadDataObjectManager(self.session), self.lock)

    def query(self, *args, **kwargs):
        #with self.lock:
        return self.session.query(*args, **kwargs)


class ExecutorPair:
    # consists in:
    # * a pair of ThreadPoolExecutor. One for transfers, the
    #   second for compute tasks.
    # * a queue to yield advancement to progressbar

    def __init__(self, session, xnt, cnt):
        self.session = LockedSession(session)
        self.xnt = xnt
        self.cnt = cnt
        self.xpe = ThreadPoolExecutor(max_workers=xnt)
        self.cpe = ThreadPoolExecutor(max_workers=cnt)

        self.interrupted = False

        # list of pending tasks
        self.xfer_futures = []
        self.compute_futures = []

        # locks for manipulating futures lists
        self.xfer_lock = Lock()
        self.compute_lock = Lock()

        # queue used to yield advancement to progressbar
        self.yield_queue = Queue()

    # context manager behaviour
    def __enter__(self):
            return self

    def __exit__(self, type, value, traceback):
        # cancel all pending xfer tasks
        for f in self.xfer_futures:
            f.cancel()

        # mark running tasks for immediate interruption
        self.interrupted = True

        # shutdown executors
        self.xpe.shutdown(wait=True)
        self.cpe.shutdown(wait=True)

    def submit_xfer(self, func, args, completed_cb=None):
        fut = self.xpe.submit(func, *args)
        return self.add_xfer_future(fut, completed_cb)

    def add_xfer_future(self, fut, completed_cb=None):
        if completed_cb is not None:
            fut.add_done_callback(completed_cb)

        with self.xfer_lock:
            self.xfer_futures.append(fut)

        return fut

    def submit_compute(self, func, args, completed_cb=None):
        fut = self.cpe.submit(func, *args)
        if completed_cb is not None:
            fut.add_done_callback(completed_cb)

        with self.compute_lock:
            self.compute_futures.append(fut)

        return fut

    def wait_xfer(self, timeout):
        with self.xfer_lock:
            completed, futures = wait(self.xfer_futures, timeout, FIRST_COMPLETED)
            self.xfer_futures = list(futures)

    def wait_compute(self, timeout):
        with self.compute_lock:
            completed, futures = wait(self.compute_futures, timeout, FIRST_COMPLETED)
            self.compute_futures = list(futures)

    def wait(self, xfer_timeout, compute_timeout):
        self.wait_xfer(xfer_timeout)
        self.wait_compute(compute_timeout)

    def pending_tasks(self):
        with self.xfer_lock, self.compute_lock:
            return len(self.xfer_futures) + len(self.compute_futures)

    def yq_put(self, element):
        self.yield_queue.put(element)

    def progress(self, osl):
        progress = 0
        while not self.yield_queue.empty():
            path, y = self.yield_queue.get()
            os = osl[path]

            if y is None:
                # file operation is finished, mark operation as done
                if os.status == catalog.OperationStatus.IN_PROGRESS:
                    os.done()
                continue

            os.progress += y
            progress += y

        return progress

    def wait_progress(self, timeout, osl):
        timeout *= 0.5

        while self.pending_tasks() > 0:
            self.wait(timeout, timeout)
            p = self.progress(osl)
            if p > 0:
                yield p

class LockedCounter:
    def __init__(self, value=0):
        self.lock = Lock()
        self.__value = value

    @property
    def value(self):
        return self.__value

    @value.setter
    def set(self, new_value):
        with self.lock:
            self.__value = new_value
            return self.__value

    def __add(self, add):
        with self.lock:
            self.__value += add
            return self.__value

    def incr(self, increment=1):
        return self.__add(increment)

    def decr(self, decrement=1):
        return self.__add(- decrement)

class iRODSCatalogBase(catalog.Catalog):
    """
    A base class Catalog for connecting to iRODS
    """

    BUFFER_SIZE = io.DEFAULT_BUFFER_SIZE * 1024

    def local_file_cksum(self, filename, algorithm=None):
        def get_digest(h):
            if h.name == 'sha256':
                return 'sha2:' + base64.b64encode(h.digest()).decode()

            return h.hexdigest()

        if algorithm is None:
            algorithm = 'md5'
            if self.session.pool.account.default_hash_scheme == 'SHA256':
                algorithm = 'sha256'

        scheme = hashlib.new(algorithm)
        with open(filename, 'rb') as f:
            for chunk in chunks(f, self.BUFFER_SIZE):
                scheme.update(chunk)
                yield len(chunk), ''

        yield 0, get_digest(scheme)

    def local_file_cksum_ref(self, filename, ref_cksum):
        algorithm = None
        if ref_cksum is None:
            algorithm = 'sha256' if self.session.pool.account.default_hash_scheme == 'SHA256' else 'md5'
        elif ref_cksum.startswith('sha2:'):
            algorithm = 'sha256'

        for y in self.local_file_cksum(filename, algorithm):
            yield y

    @classmethod
    def encode(cls, s):
        return password_obfuscation.encode(s, _getuid())

    @classmethod
    def decode(cls, s):
        return password_obfuscation.decode(s, _getuid())

    def __init__(self, session, default_resc, local_checksum,
                 xfer_nt=1):
        self.session = session

        self.default_resc = default_resc

        self.local_checksum = local_checksum

        self.dom = MultithreadDataObjectManager(self.session)
        self.cm = self.session.collections
        self.am = self.session.permissions
        self.xfer_nt = xfer_nt

    def close(self):
        self.session.cleanup()

    def cksum_factor(self):
        return 2 if self.local_checksum else 1

    def splitname(self, path):
        return path.rsplit('/', 1)

    def basename(self, path):
        _, basename = self.splitname(path)
        return basename

    def dirname(self, path):
        dirname, _ = self.splitname(path)
        return dirname

    def normpath(self, path):
        pathlist = []

        for t in path.split('/'):
            if t == '' or t == '.':
                continue

            if t == '..':
                if len(pathlist):
                    del pathlist[-1]
                continue

            pathlist.append(t)

        normpath = '/'.join([''] + pathlist)

        return normpath or '/'

    @method_translate_exceptions
    def lstat(self, path):
        if self.isdir(path):
            return self.lstat_dir(path)
        return self.lstat_file(path)

    def lstat_dir(self, path):
        q = self.session.query(Collection.owner_name)
        q = q.filter(Collection.name == path)

        r = q.one()

        ret = {
            'user': r[Collection.owner_name],
            'size': '',
            'mtime': '',
            'nreplicas': '',
            'isdir': True,
        }

        return ret

    def lstat_dirs(self, parent_path):
        q = self.session.query(Collection.name, Collection.owner_name)
        q = q.filter(Collection.parent_name == parent_path)

        ret = {}
        for r in q.get_results():

            ret[self.basename(r[Collection.name])] = {
                'user': r[Collection.owner_name],
                'size': '',
                'mtime': '',
                'nreplicas': '',
                'isdir': True,
            }

        return ret

    def lstat_files(self, dirname):
        epoch = datetime.datetime(1, 1, 1)

        q = self.session.query(DataObject.name, DataObject.owner_name,
                               DataObject.size,
                               DataObject.modify_time,
                               DataObject.replica_number)
        q = q.filter(Collection.name == dirname)

        ret = {}
        for r in q.get_results():
            name = r[DataObject.name]
            dobj = ret.get(name, {'minsize': None, 'maxsize': 0,
                                  'mtime': epoch, 'isdir': False,
                                  'nreplicas': 0})

            dobj['user'] = r[DataObject.owner_name]
            dobj['nreplicas'] += 1

            mtime = r[DataObject.modify_time]
            if mtime > dobj['mtime']:
                dobj['mtime'] = mtime

            size = r[DataObject.size]
            if dobj['minsize'] is None or size < dobj['minsize']:
                dobj['minsize'] = size

            if size > dobj['maxsize']:
                dobj['maxsize'] = size

            ret[name] = dobj

        for k, v in ret.items():
            minsize = v['minsize']
            maxsize = v['maxsize']
            if maxsize != minsize:
                v['size'] = '{}-{}'.format(minsize, maxsize)
            else:
                v['size'] = str(minsize)

        return ret

    def lstat_file(self, path):
        dirname, basename = self.splitname(path)
        q = self.session.query(DataObject.owner_name, DataObject.size,
                               DataObject.modify_time,
                               DataObject.replica_number)
        q = q.filter(Collection.name == dirname)
        q = q.filter(DataObject.name == basename)

        replicas = q.all()
        if len(replicas) < 1:
            # no replica
            raise exceptions.ioerror(exceptions.errno.ENOENT)

        ret = {'nreplicas': len(replicas), 'isdir': False}
        minsize = None
        maxsize = 0
        for r in replicas:
            ret['user'] = r[DataObject.owner_name]
            ret['mtime'] = r[DataObject.modify_time]
            size = r[DataObject.size]
            if minsize is None or size < minsize:
                minsize = size
            if size > maxsize:
                maxsize = size

        if maxsize != minsize:
            ret['size'] = '{}-{}'.format(minsize, maxsize)
        else:
            ret['size'] = str(minsize)

        return ret

    @method_translate_exceptions
    def listdir(self, path):
        q = self.session.query(DataObject.name).filter(Collection.name == path)
        files = [r[DataObject.name] for r in q.all()]

        q = self.session.query(Collection.name) \
            .filter(Collection.parent_name == path)
        colls = [self.basename(c[Collection.name]) for c in q.all()]

        ret = self.lstat_dirs(path)
        ret.update(self.lstat_files(path))

        return ret

    @method_translate_exceptions
    def isdir(self, path):
        q = self.session.query(Collection.id).filter(Collection.name == path)

        try:
            q.one()
        except irods.exception.NoResultFound:
            return False

        return True

    def join(self, *args):
        return '/'.join(args)

    def remote_files_stats(self, file_paths):
        dirs = {self.dirname(p) for p in file_paths}

        stats = {}
        for d in dirs:
            q = self.session.query(DataObject.name, DataObject.size)
            q = q.filter(Collection.name == d)

            for r in q.get_results():
                stats[self.join(d, r[DataObject.name])] = \
                  int(r[DataObject.size])

        return (len(file_paths),
                sum([v for k, v in stats.items() if k in file_paths]),
                stats)

    def remote_trees_stats(self, dirs):
        nfiles = 0
        size = 0
        stats = {}

        for d in dirs:
            # need to keep column 'collection_id' to avoid 'distinct' clause on
            # recursive queries
            q = self.session.query(DataObject.collection_id, DataObject.name,
                                   DataObject.size)

            dsize = 0
            # first level query
            q1 = q.filter(Collection.name == d)
            for r in q1.get_results():
                nfiles += 1
                dsize += int(r[DataObject.size])

            # recursive query
            qr = q.filter(Like(Collection.name, self.join(d, '%')))
            for r in qr.get_results():
                nfiles += 1
                dsize += int(r[DataObject.size])

            stats[d] = dsize
            size += dsize

        return nfiles, size, stats

    def _register_data_objects_in_osl(self, pathlist, osl):
        def _cancel_cb(op_stat):
            print_('Operation', op_stat.print_status(),
                   ': delete', op_stat.current_element)
            os.unlink(op_stat.current_element)

        nfiles, file_size, fstats = self.remote_files_stats(pathlist)

        cksum_factor = self.cksum_factor()
        stats = {k : (s * cksum_factor) for k, s in fstats.items()}
        osl.update_list(pathlist, size=stats, file_size=fstats,
                        cancel=_cancel_cb)


        return nfiles, file_size, file_size * cksum_factor

    @method_translate_exceptions
    def download_files(self, pathlist, destdir, osl):
        t0 = time.time()

        nfiles, file_size, size = self._register_data_objects_in_osl(pathlist, osl)

        if nfiles > 1 or size > self.BUFFER_SIZE:
            # wake up progress bar for more than one file or one large file
            yield 0, size

        completed = 0
        with ExecutorPair(self.session, self.xfer_nt, cnt=4) as ep:
            for y in self._download_files(pathlist, destdir, osl, ep):
                completed += y
                yield completed, size

            # actively wait for all tasks to finish
            for p in ep.wait_progress(0.1, osl):
                completed += p
                yield completed, size

        elapsed = time.time() - t0
        if elapsed > 0:
            print_('download_files finished', elapsed, 'seconds (%.2f MB/s)' % (file_size / (elapsed * 10**6)))

    def _download_files(self, irods_paths, local_dir, osl, ep):
        # local functions
        # ---------------
        def _submit_read(ep, irods_path, local_path, **options):
            @function_translate_exceptions
            def _read(p, s, task_num):
                f = dom.open(irods_path, 'r', **options)
                fd = os.open(local_path, os.O_WRONLY, 0o640)

                try:
                    buf = bytearray(self.BUFFER_SIZE)
                    read = 0
                    origp = p
                    #print_('thread if={} p={} s={} fd={}'.format(irods_path, p, s, fd))
                    f.seek(p)
                    os.lseek(fd, p, os.SEEK_SET)
                    while s > 0:
                        if ep.interrupted:
                            return

                        to_read = min(self.BUFFER_SIZE, s)
                        mv = memoryview(buf)[0:to_read]
                        l = f.readinto(mv)
                        if l != to_read:
                            raise IOError('[%s->%s] error reading %s expected %s' % (irods_path, local_path, l, to_read))

                        os.write(fd, mv)
                        p += l
                        s -= l
                        read += l

                        ep.yq_put((irods_path, l))
                        #print_('%s read %s, remaining %s' % (origp, read, s))

                finally:
                    f.close()
                    os.close(fd)

                return irods_path, local_path, read

            dom = ep.session.data_objects

            fsize = osl[irods_path].file_size

            with open(local_path, 'wb+') as lf:
                lf.truncate(fsize)

            i = 0
            futures = []
            for p, s in splitsize(fsize, ep.xnt, self.BUFFER_SIZE):
                print_(time.time(), 'submit read', irods_path, p, s, fsize)
                fut = ep.submit_xfer(_read, (p, s, i), _submit_validate)
                futures.append(fut)
                i += 1

            return futures

        def progress_cb(local_path, irods_path, count):
            print_('progress', local_path, irods_path, count)
            ep.yq_put((irods_path, count))

        def _submit_validate(fut):
            # validate downloaded file (from xfer future fut)

            @function_translate_exceptions
            def _validate(irods_path, local_path):
                print_(time.time(), 'validating file', local_path)
                obj = self.dom.get(irods_path)

                completed = 0

                for l, local_cksum in self.local_file_cksum_ref(local_path, obj.checksum):
                    if ep.interrupted:
                        return

                    completed += l

                    ep.yq_put((irods_path, l))

                if obj.checksum is None:
                    print_('WARNING: iRODS checksum is None', irods_path, local_cksum)
                elif obj.checksum != local_cksum:
                    msg = 'Downloaded file {} has an incorrect checksum ' \
                           '(local=\'{}\', ' \
                           'catalog=\'{}\')'.format(irods_path, local_cksum, obj.checksum)
                    osl[irods_path].fail(exceptions.ChecksumError(msg))
                else:
                    print_('checksum ok', irods_path, local_cksum)

                # tell file validation is finished
                ep.yq_put((irods_path, None))

            if fut.cancelled() or ep.interrupted:
                # bailout since result is not available
                return

            op_stat = fut2os[fut]
            del fut2os[fut]

            e = fut.exception()
            if e is not None:
                print_('ERROR: caught some exception', e)
                op_stat.fail(e)
                return

            #ip, lp, read_bytes = fut.result()
            ip, lp = fut2files[fut]
            files[(ip, lp)] -= 1

            # whole file downloaded: submit validation task
            if files[(ip, lp)] == 0:
                if not self.local_checksum:
                    # checksum not needed: tell file validation is finished
                    ep.yq_put((ip, None))
                    return

                vfut = ep.submit_compute(_validate, (ip, lp), fut_handle_exception)
                fut2os[vfut] = op_stat

        def fut_handle_exception(fut):
            e = fut.exception()
            if e is not None:
                print_('ERROR: got some exception', e)
                fut2os[fut].fail(e)

        # method code
        # -----------

        options = {
            kw.FORCE_FLAG_KW: True,
        }

        files = {}
        fut2os = {}
        fut2files = {}

        # submit transfer tasks
        for irods_path in irods_paths:
            _, fn = irods_path.rsplit('/', 1)
            local_path = os.path.join(local_dir, fn)

            osl[irods_path].in_progress(local_path)
            print_('in progress', irods_path)
#            futs = _submit_read(ep, irods_path, local_path, **options)
            futs = self.dom.download_parallel(irods_path, local_path, ep.xpe,
                                              progress_cb=progress_cb, **options)
            for fut in futs:
                ep.add_xfer_future(fut, completed_cb=_submit_validate)
                fut2files[fut] = irods_path, local_path

            files[(irods_path, local_path)] = len(futs)
            fut2os.update({fut : osl[irods_path] for fut in futs})

            # check already finished transfer tasks
            ep.wait_xfer(0)
            ep.wait_compute(0)
            p = ep.progress(osl)
            if p > 0:
                yield p
                
        yield ep.progress(osl)

    def _download_coll(self, coll, destdir, osl, ep):
        destdir = os.path.join(destdir, coll.name)
        try:
            os.makedirs(destdir)
        except OSError:
            if not os.path.isdir(destdir):
                raise

        print_('begin coll', coll.name, 'subcolls', list(coll.subcollections))

        pathlist = [dobj.path for dobj in list(coll.data_objects)]
        osl.add_keys(pathlist)
        self._register_data_objects_in_osl(pathlist, osl)
        for y in self._download_files(pathlist, destdir, osl, ep):
            yield y

        for subcoll in coll.subcollections:
            for y in self._download_coll(subcoll, destdir, osl, ep):
                yield y

    @method_translate_exceptions
    def download_directories(self, pathlist, destdir, osl):
        t0 = time.time()

        nfiles, file_size, fstats = self.remote_trees_stats(pathlist)
        cksum_factor = self.cksum_factor()
        size = file_size * cksum_factor

        with ExecutorPair(self.session, self.xfer_nt, cnt=4) as ep:
            completed = 0
            for p in pathlist:
                coll = self.cm.get(p)
                for y in self._download_coll(coll, destdir, osl, ep):
                    completed += y
                    yield completed, size

            # actively wait for all tasks to finish
            for p in ep.wait_progress(0.1, osl):
                completed += p
                yield completed, size

        elapsed = time.time() - t0
        if elapsed > 0:
            print_('download_directories finished', elapsed, 'seconds (%.2f MB/s)' % (file_size / (elapsed * 10**6)))

    @method_translate_exceptions
    def upload_files(self, files, path, osl):
        t0 = time.time()

        nfiles, fsize, fstats = local_files_stats(files)

        def cancel(op_stat):
            print_('Operation', op_stat.print_status(),
                   ': delete', op_stat.current_element)
            try:
                self.dom.unlink(op_stat.current_element, force=True)
            except irods.exception.CAT_NO_ROWS_FOUND:
                # don't care if file is not found: catalog may have
                # removed it already
                pass

        cksum_factor = self.cksum_factor()
        size = fsize * cksum_factor
        stats = {k : (s * cksum_factor) for k, s in fstats.items()}
        osl.update_list(files, size=stats, file_size=fstats, cancel=cancel)

        with ExecutorPair(self.session, self.xfer_nt, cnt=4) as ep:
            completed = 0
            for s in self._upload_files(files, path, osl, ep):
                completed += s
                yield completed, size

            # actively wait for all tasks to finish
            for p in ep.wait_progress(0.1, osl):
                completed += p
                yield completed, size

        elapsed = time.time() - t0
        if elapsed > 0:
            print_('upload_files finished in', elapsed,
                   'seconds (%.2f MB/s)' % (fsize/elapsed / 10**6))

    def _upload_files(self, files, path, osl, ep):
        def _upload(local_path, irods_path, op_stat, p, s, lc, cksum,
                    options):
            with open(local_path, 'rb') as f:
                dom = ep.session.data_objects
                o = dom.open(irods_path, 'a+', **options)

                f.seek(p)
                o.seek(p)

                completed = 0
                while completed < s:
                    if ep.interrupted:
                        break
                    chunk = f.read(self.BUFFER_SIZE)
                    l = len(chunk)
                    o.write(chunk)
                    completed += l
                    ep.yq_put((local_path, l))

                o.close()


            # we are the last upload task
            if lc.decr() == 0:
                print_(time.time(), 'ok xfer finished', local_path, p, p+s, cksum)

                opts = options.copy()

                # Set operation type to trigger acPostProcForPut
                opts[kw.OPR_TYPE_KW] = 1  # PUT_OPR
                if cksum is not None:
                    opts[kw.VERIFY_CHKSUM_KW] = cksum

                if ep.interrupted:
                    return

                o = dom.open(irods_path, 'a+', **opts)
                o.close()

                if kw.ALL_KW in opts:
                    opts[kw.UPDATE_REPL_KW] = ''

                    if ep.interrupted:
                        return

                    dom.replicate(irods_path, **opts)

                if ep.interrupted:
                    return

                ep.yq_put((local_path, None))

        def submit_compute_checksum(lf, op_stat):
            # variables lf and op_stat are used in local functions
            def _compute_checksum():
                def _upload_completed_cb(fut):
                    if fut.cancelled():
                        return

                    e = fut.exception()
                    if e is not None:
                        print_(lf, '_upload failed with exception', e)
                        op_stat.fail(e)
                        return

                print_(time.time(), 'begin _compute_checksum', lf)

                cksum = None
                if self.local_checksum:
                    l = 0
                    for l, cksum in self.local_file_cksum(lf):
                        ep.yq_put((lf, l))

                try:
                    # overwrite an existing file in parallel seems to be
                    # a bad idea, so delete it if necessary
                    ep.session.data_objects.unlink(op_stat.current_element)
                except irods.exception.CAT_NO_ROWS_FOUND:
                    pass

                file_parts = list(splitsize(op_stat.file_size, ep.xnt,
                                  self.BUFFER_SIZE))

                xfer_counter = LockedCounter(len(file_parts))

                if ep.interrupted:
                    return

                for p, s in file_parts:
                    print_(time.time(), 'submit _upload', lf, p, p + s, op_stat.file_size)
                    ep.submit_xfer(_upload, (lf, op_stat.current_element,
                                             op_stat, p, s, xfer_counter,
                                             cksum, options),
                                   _upload_completed_cb)

            def _completed_cb(fut):
                if fut.cancelled():
                    return

                e = fut.exception()
                if e is not None:
                    print_(lf, 'compute checksum failed with exception', e)
                    op_stat.fail(e)
                    return

            print_(time.time(), 'submit _compute_checksum', lf)
            fut = ep.submit_compute(_compute_checksum, [], _completed_cb)

        # method code
        #############
        if not path.endswith('/'):
            path = path + '/'

        options = {
            kw.ALL_KW: '',
            kw.UPDATE_REPL_KW: '',
        }

        if self.default_resc is not None:
            options[kw.DEST_RESC_NAME_KW] = self.default_resc

        for f in files:
            basename = os.path.basename(f)
            irods_path = path + basename

            osl[f].in_progress(irods_path)

            submit_compute_checksum(f, osl[f])

            ep.wait(0, 0)
            y = ep.progress(osl)
            yield y

    def _upload_dir(self, dir_, path, osl, ep):
        print_(time.time(), '_upload_dir begin', dir_)
        files = []
        subdirs = []

        try:
            self.cm.create(path)
        except irods.exception.CATALOG_ALREADY_HAS_ITEM_BY_THAT_NAME:
            pass

        for name in os.listdir(dir_):
            abspath = os.path.join(dir_, name)
            if os.path.isdir(abspath):
                subdirs.append((abspath, name))
            else:
                files.append(abspath)

        nfiles, fsize, fstats = local_files_stats(files)

        def cancel(op_stat):
            print_('Operation', op_stat.print_status(),
                   ': delete', op_stat.current_element)
            try:
                self.dom.unlink(op_stat.current_element, force=True)
            except irods.exception.CAT_NO_ROWS_FOUND:
                # don't care if file is not found: catalog may have
                # removed it already
                pass

        cksum_factor = self.cksum_factor()
        stats = {k : (s * cksum_factor) for k, s in fstats.items()}
        osl.add_keys(files)
        osl.update_list(files, size=stats, file_size=fstats, cancel=cancel)

        for abspath, name in subdirs:
            cpath = self.join(path, name)

            for y in self._upload_dir(abspath, cpath, osl, ep):
                yield y

        for y in self._upload_files(files, path, osl, ep):
            yield y

        print_(time.time(), '_upload_dir end', dir_)

    @method_translate_exceptions
    def upload_directories(self, dirs, path, osl):
        t0 = time.time()
        print_(t0, 'upload_directories', dirs)
        nfiles, fsize, stats = local_trees_stats(dirs)

        cksum_factor = self.cksum_factor()
        size = fsize * cksum_factor

        with ExecutorPair(self.session, self.xfer_nt, cnt=4) as ep:
            completed = 0
            for d in dirs:
                name = os.path.basename(d)
                cpath = self.join(path, name)

                for s in self._upload_dir(d, cpath, osl, ep):
                    completed += s
                    yield completed, size

            # actively wait for all tasks to finish
            for p in ep.wait_progress(0.1, osl):
                completed += p
                yield completed, size

        elapsed = time.time() - t0
        if elapsed > 0:
            print_('upload_directories finished in', elapsed,
                   'seconds (%.2f MB/s)' % (fsize/elapsed / 10**6))

    @method_translate_exceptions
    def delete_files(self, files, osl):
        number = len(files)

        i = 0
        for f in files:
            osl[f].size=1
            osl[f].in_progress(None)
            self.dom.unlink(f, force=True)
            osl[f].done()
            i += 1
            yield i, number

    def _coll_remove_yield(self, path, recurse=True, force=False, **options):
        """
        interruptible version of CollectionManager.coll_remove() method
        """
        if recurse:
            options[kw.RECURSIVE_OPR__KW] = ''
        if force:
            options[kw.FORCE_FLAG_KW] = ''

        try:
            oprType = options[kw.OPR_TYPE_KW]
        except KeyError:
            oprType = 0

        message_body = message.CollectionRequest(
            collName=path,
            flags = 0,
            oprType = oprType,
            KeyValPair_PI=message.StringStringMap(options)
        )
        msg = message.iRODSMessage('RODS_API_REQ', msg=message_body,
                                   int_info=api_number['RM_COLL_AN'])
        with self.session.pool.get_connection() as conn:
            conn.send(msg)
            response = conn.recv()

            try:
                while response.int_info == const.SYS_SVR_TO_CLI_COLL_STAT:
                    conn.reply(const.SYS_CLI_TO_SVR_COLL_STAT_REPLY)
                    yield
                    response = conn.recv()
            except GeneratorExit:
                # destroy connection which is in a bad state (could fix?)
                conn.release(destroy=True)

    @method_translate_exceptions
    def delete_directories(self, directories, osl):
        number = len(directories)

        i = 0
        for d in directories:
            osl[d].size=1
            osl[d].in_progress(None)
            for y in self._coll_remove_yield(d, recurse=True, force=True):
                yield i, number
            osl[d].done()
            i += 1
            yield i, number

    @method_translate_exceptions
    def mkdir(self, path):
        self.cm.create(path)

    def __acls_from_object(self, obj):
        access = self.am.get(obj)

        acls = [a.__dict__.copy() for a in access]

        for a in acls:
            # build a unique ttk.TreeWidget iid
            a['iid'] = '#'.join([a[t] for t in ['user_name', 'user_zone',
                                                'access_name']])
            a['#0'] = a['user_name']

        @function_translate_exceptions
        def add(result):
            access_name = result['access_name']
            user_name = result['#0']
            user_zone = result['user_zone']

            acl = irods.access.iRODSAccess(access_name, obj.path, user_name,
                                           user_zone)

            self.session.permissions.set(acl)

            return '#'.join([user_name, user_zone, access_name])

        @function_translate_exceptions
        def remove(result):
            user_name = result['#0']
            user_zone = result['user_zone']

            acl = irods.access.iRODSAccess('null', obj.path, user_name,
                                           user_zone)

            self.session.permissions.set(acl)

        acls_def = iRODSCatalogBase.acls_def(self.session.zone)

        return List(acls_def, acls, add_cb=add, remove_cb=remove)

    def __metadata_from_object(self, obj):
        metadata = [md.__dict__.copy() for md in obj.metadata.items()]

        for md in metadata:
            md['iid'] = md['avu_id']
            md['#0'] = md['name']

        @function_translate_exceptions
        def add(result):
            name = result['#0']
            value = result['value']
            units = result['units']

            obj.metadata.add(name, value, units)

            for md in obj.metadata.items():
                if md.name == name and md.value and md.units == units:
                    return md.avu_id

            return None

        @function_translate_exceptions
        def remove(result):
            name = result['#0']
            value = result['value']
            unit = result['units']

            obj.metadata.remove(name, value, unit)

        return List(iRODSCatalogBase.metadata_def(), metadata, add_cb=add,
                    remove_cb=remove)

    @method_translate_exceptions
    def directory_properties(self, path):
        co = self.cm.get(path)
        acls_list = self.__acls_from_object(co)
        metadata_list = self.__metadata_from_object(co)

        inheritance = self.session.query(Collection.inheritance) \
            .filter(Collection.name == path).one()[Collection.inheritance] \
            == '1'

        def inheritance_changed(value):
            name = 'inherit' if value else 'noinherit'
            acl = irods.access.iRODSAccess(name, path, '', '')

            self.session.permissions.set(acl)

        f = form.BooleanField('Inherit:', inheritance,
                              state_change_cb=inheritance_changed)
        inherit_frame = form.FrameGenerator([f])

        return collections.OrderedDict([
            ('Permissions', acls_list),
            ('Metadata', metadata_list),
            ('Inheritance', inherit_frame),
        ])

    @method_translate_exceptions
    def file_properties(self, path):
        do = self.dom.get(path)
        replicas = [r.__dict__.copy() for r in do.replicas]
        for r in replicas:
            # set row title for ListManager use
            r['#0'] = r['number']

        replicas_list = List(iRODSCatalogBase.replicas_def(), replicas)

        acls_list = self.__acls_from_object(do)
        metadata_list = self.__metadata_from_object(do)

        return collections.OrderedDict([
            ('Replicas', replicas_list),
            ('Permissions', acls_list),
            ('Metadata', metadata_list),
        ])

    @classmethod
    def replicas_def(cls):
        repl_num = ColumnDef('#0', 'Number',
                             form_field=form.IntegerField('Replica number:',
                                                          -1))
        resc = ColumnDef('resource_name', 'Resource',
                         form_field=form.TextField('Resource name:'))
        path = ColumnDef('path', 'Path',
                         form_field=form.TextField('Replica path:'))
        status = ColumnDef('status', 'Status',
                           form_field=form.TextField('Replica status:'))
        checksum = ColumnDef('checksum', 'Checksum',
                             form_field=form.TextField('Replica checksum:'))

        cols = [repl_num, resc, status, checksum, path]

        return collections.OrderedDict([(cd.name, cd) for cd in cols])

    @classmethod
    def acls_def(cls, default_zone=''):
        user = ColumnDef('#0', 'User', form_field=form.TextField('User:'))
        zone = ColumnDef('user_zone', 'Zone',
                         form_field=form.TextField('User zone:', default_zone))
        access_types = ['read', 'write', 'own']
        type = ColumnDef('access_name', 'Access type',
                         form_field=form.ComboboxChoiceField('Acces type:',
                                                             access_types,
                                                             access_types[0]))

        cols = [user, zone, type]

        return collections.OrderedDict([(cd.name, cd) for cd in cols])

    @classmethod
    def metadata_def(cls):
        name = ColumnDef('#0', 'Name',
                         form_field=form.TextField('Metadata name:'))
        value = ColumnDef('value', 'Value',
                          form_field=form.TextField('Metadata value:'))
        unit = ColumnDef('units', 'Unit',
                         form_field=form.TextField('Metadata unit:'))

        cols = [name, value, unit]

        return collections.OrderedDict([(cd.name, cd) for cd in cols])

    @classmethod
    def config_fields(cls):

        tags = ['inline_config']

        return collections.OrderedDict([
            ('local_checksum', form.BooleanField('Perform local checksum:',
                                                 default_value=True)),
            ('max_xfer_threads',
             form.IntegerField('Maximum number of transfer threads:',
                               default_value='16')),
            ('use_irods_env', form.BooleanField('Use irods environment file:',
                                                disables_tags=tags)),
            ('host', form.HostnameField('iRODS host:', tags=tags)),
            ('port', form.IntegerField('iRODS port:', '1247', tags=tags)),
            ('zone', form.TextField('iRODS zone:', tags=tags)),
            ('user_name', form.TextField('iRODS user name:', tags=tags)),
            ('default_resc', form.TextField('Default resource:', tags=tags)),
            ('store_password', form.BooleanField('Remember password:',
                                                 enables_tags=['password'],
                                                 tags=tags)),
            ('password', form.PasswordField('iRODS password:',
                                            encode=cls.encode,
                                            decode=cls.decode,
                                            tags=tags + ['password'])),
        ])


class iRODSCatalog3(iRODSCatalogBase):
    """
    A Catalog for connecting to iRODS v3
    """

    def __init__(self, host, port, user, zone, scrambled_password,
                 default_resc, local_checksum, xfer_nt):
        try:
            password = iRODSCatalogBase.decode(scrambled_password)
            session = iRODSSession(host=host, port=port, user=user,
                                   password=password, zone=zone,
                                   default_hash_scheme='MD5')
        except irods.exception.CAT_INVALID_AUTHENTICATION as e:
            raise exceptions.ConnectionError(e)

        super(iRODSCatalog3, self).__init__(session, default_resc,
                                            local_checksum, xfer_nt)


class iRODSCatalog4(iRODSCatalogBase):
    """
    A Catalog for connecting to iRODS v4
    """

    @classmethod
    def from_env_file(cls, env_file, local_checksum):
        session = iRODSSession(irods_env_file=env_file)

        return cls(session, None, local_checksum, 16)

    @classmethod
    def from_options(cls, host, port, user, zone, scrambled_password,
                     default_resc, local_checksum, default_hash_scheme,
                     xfer_nt, ssl_settings=None):
        kwargs = {}
        try:
            password = iRODSCatalogBase.decode(scrambled_password)
            kwargs.update(dict(host=host, port=port, user=user,
                               password=password, zone=zone,
                               default_hash_scheme=default_hash_scheme))
        except irods.exception.CAT_INVALID_AUTHENTICATION as e:
            raise exceptions.ConnectionError(e)

        if ssl_settings is not None:
            kwargs.update(ssl_settings)

        session = iRODSSession(**kwargs)

        return cls(session, default_resc, local_checksum, xfer_nt)

    @classmethod
    def config_fields(cls):
        base_dict = iRODSCatalogBase.config_fields()

        tags = base_dict['host'].tags

        base_dict.update({
            ('irods_default_hash_scheme',
             form.RadioChoiceField('Default hash scheme',
                                   values=['MD5', 'SHA256'],
                                   default_value='SHA256',
                                   tags=tags)),
        })

        ssl_tag = 'ssl_config'

        ssl_tags = tags + [ssl_tag]

        ssl_options = collections.OrderedDict([
            ('irods_client_server_policy',
             form.RadioChoiceField('irods_client_server_policy',
                                   values=['CS_NEG_REQUIRE', 'CS_NEG_REFUSE',
                                           'CS_NEG_DONT_CARE'],
                                   default_value='CS_NEG_REQUIRE',
                                   tags=tags)),

            ('use_irods_ssl', form.BooleanField('Use irods SSL transfer',
                                                enables_tags=[ssl_tag],
                                                tags=tags)),
            ('irods_encryption_algorithm',
             form.TextField('irods_encryption_algorithm:',
                            default_value='AES-256-CBC', tags=ssl_tags)),
            ('irods_encryption_key_size',
             form.IntegerField('irods_encryption_key_size:', default_value='32',
                               tags=ssl_tags)),
            ('irods_encryption_num_hash_rounds',
             form.IntegerField('irods_encryption_num_hash_rounds:',
                               default_value='16', tags=ssl_tags)),
            ('irods_encryption_salt_size',
             form.IntegerField('irods_encryption_salt_size:', default_value='8',
                               tags=ssl_tags)),
            ('irods_ssl_ca_certificate_file',
             form.FileSelectorField('irods_ssl_ca_certificate_file:',
                                    tags=ssl_tags)),
        ])

        base_dict.update(ssl_options)

        return base_dict


def irods3_catalog_from_envfile(envfile, local_checksum, xfer_nt):
    """
    Creates an iRODSCatalog from a iRODS v3 configuration file (like
    "~/.irods/.irodsEnv")
    """
    env3 = parse_env3(envfile)

    host = env3['irodsHost']
    port = int(env3.get('irodsPort', '1247'))
    user = env3['irodsUserName']
    zone = env3['irodsZone']
    pwdfile = env3['irodsAuthFileName']
    default_resc = env3.get('irodsDefResource', None)

    with open(pwdfile, 'r') as f:
        scrambled_password = f.read().strip()

    return iRODSCatalog3(host, port, user, zone, scrambled_password,
                         default_resc, local_checksum, xfer_nt)


def irods3_catalog_from_config(cfg):
    """
    Creates an iRODSCatalog from configuration
    """
    local_checksum = option_is_true(cfg.get('local_checksum', 'True'))
    xfer_nt = int(cfg.get('max_xfer_threads', '16'))
    use_env = option_is_true(cfg['use_irods_env'])

    if use_env:
        envfile = os.path.join(os.path.expanduser('~'), '.irods', '.irodsEnv')
        return lambda master: irods3_catalog_from_envfile(envfile,
                                                          local_checksum, xfer_nt)

    host = cfg['host']
    port = cfg['port']
    user = cfg['user_name']
    zone = cfg['zone']

    default_resc = cfg.get('default_resc', None)
    if default_resc == '':
        # empty string is not valid for default_resc
        default_resc = None

    store_password = cfg['store_password']
    scrambled_password = None
    if option_is_true(store_password):
        scrambled_password = cfg['password']
        return lambda master: iRODSCatalog3(host, port, user, zone,
                                            scrambled_password, default_resc,
                                            local_checksum, xfer_nt)
    else:
        def ask_password(master):
            cancelled = {'cancelled': False}

            def _do_ok(e=None):
                tl.destroy()

            def _do_cancel(e=None):
                pf.from_string('')
                cancelled['cancelled'] = True
                tl.destroy()

            tl = tk.Toplevel(master)
            tl.title('iRODS password')
            tl.transient(master)

            ff = form.FormFrame(tl)
            pf = form.PasswordField('password for {}@{}:'.format(user, zone),
                                    return_cb=_do_ok)
            ff.grid_fields([pf])
            ff.pack()

            butbox = tk.Frame(tl)
            butbox.pack()
            ok = tk.Button(butbox, text='Ok', command=_do_ok)
            ok.grid()
            ok.bind('<Return>', _do_ok)
            cancel = tk.Button(butbox, text='Cancel', command=_do_cancel)
            cancel.grid(row=0, column=1)
            cancel.bind('<Return>', _do_cancel)

            tl.wait_window()

            if cancelled['cancelled']:
                return None

            scrambled_password = iRODSCatalogBase.encode(pf.to_string())

            return iRODSCatalog3(host, port, user, zone, scrambled_password,
                                 default_resc, local_checksum, xfer_nt)

        return ask_password


def irods4_catalog_from_config(cfg):
    """
    Creates an iRODSCatalog from configuration
    """
    local_checksum = option_is_true(cfg.get('local_checksum', 'True'))
    xfer_nt = int(cfg.get('max_xfer_threads', '16'))
    use_env = option_is_true(cfg['use_irods_env'])

    if use_env:
        envfile = os.path.join(os.path.expanduser('~'), '.irods',
                               'irods_environment.json')
        return lambda master: iRODSCatalog4.from_env_file(envfile,
                                                          local_checksum)

    host = cfg['host']
    port = cfg['port']
    user = cfg['user_name']
    zone = cfg['zone']

    default_hash_scheme = cfg['irods_default_hash_scheme']

    ssl = None
    if option_is_true(cfg['use_irods_ssl']):
        ssl = {
            'irods_client_server_negotiation': 'request_server_negotiation',
            'irods_client_server_policy': cfg['irods_client_server_policy'],
            'irods_encryption_algorithm': cfg['irods_encryption_algorithm'],
            'irods_encryption_key_size': int(cfg['irods_encryption_key_size']),
            'irods_encryption_num_hash_rounds':
                int(cfg['irods_encryption_num_hash_rounds']),
            'irods_encryption_salt_size':
                int(cfg['irods_encryption_salt_size']),
            'irods_ssl_ca_certificate_file':
                cfg['irods_ssl_ca_certificate_file'],
        }

    default_resc = cfg.get('default_resc', None)
    if default_resc == '':
        # empty string is not valid for default_resc
        default_resc = None

    store_password = cfg['store_password']
    scrambled_password = None
    if option_is_true(store_password):
        scrambled_password = cfg['password']
        return lambda master: iRODSCatalog4.from_options(host, port, user,
                                                         zone,
                                                         scrambled_password,
                                                         default_resc,
                                                         local_checksum,
                                                         default_hash_scheme,
                                                         xfer_nt,
                                                         ssl)
    else:
        def ask_password(master):
            cancelled = {'cancelled': False}

            def _do_ok(e=None):
                tl.destroy()

            def _do_cancel(e=None):
                pf.from_string('')
                cancelled['cancelled'] = True
                tl.destroy()

            tl = tk.Toplevel(master)
            tl.title('iRODS password')
            tl.transient(master)

            ff = form.FormFrame(tl)
            pf = form.PasswordField('password for {}@{}:'.format(user, zone),
                                    return_cb=_do_ok)
            ff.grid_fields([pf])
            ff.pack()

            butbox = tk.Frame(tl)
            butbox.pack()
            ok = tk.Button(butbox, text='Ok', command=_do_ok)
            ok.grid()
            ok.bind('<Return>', _do_ok)
            cancel = tk.Button(butbox, text='Cancel', command=_do_cancel)
            cancel.grid(row=0, column=1)
            cancel.bind('<Return>', _do_cancel)

            tl.wait_window()

            if cancelled['cancelled']:
                return None

            scrambled_password = iRODSCatalogBase.encode(pf.to_string())

            return iRODSCatalog4.from_options(host, port, user, zone,
                                              scrambled_password, default_resc,
                                              local_checksum,
                                              default_hash_scheme,
                                              xfer_nt, ssl)

        return ask_password



"""
Navigation bar
"""

from six import print_
from six.moves import tkinter as tk
from six.moves import tkinter_ttk as ttk


class HistoryCombobox(ttk.Combobox):
    def __init__(self, master, initial_path):
        ttk.Combobox.__init__(self, master)

        self.history = []
        if initial_path:
            self.delete(0, tk.END)
            self.insert(0, initial_path)
            self.history_append(initial_path)


    def history_append(self, value):
        if value in self.history:
            self.history.remove(value)
        self.history.append(value)
        self.config(values=self.history)

    def history_limit(self, keep=0):
        self.history = self.history[0:keep]
        self.config(values=self.history)


class NavigationBar(tk.Frame):
    def __init__(self, master, initial_path='', change_path_cb = None):
        tk.Frame.__init__(self, master)

        self.change_path_cb = change_path_cb

        self.refresh_but = tk.Button(self, text='Refresh')
        self.refresh_but.grid(row=0, sticky='w')

        self.path_entry = HistoryCombobox(self, initial_path)
        self.path_entry.grid(row=0, column=1, sticky='ew')
        self.path_entry.bind('<Return>', self.path_changed)
        self.path_entry.bind('<<ComboboxSelected>>', self.path_changed)

        self.columnconfigure(1, weight=1)

    def get_path(self):
        return self.path_entry.get()

    def set_path(self, new_path, clear_history=False):
        if clear_history:
            self.path_entry.history_limit(0)
        self.path_entry.set(new_path)
        self.path_entry.history_append(new_path)

    def path_changed(self, e):
        print_('new path', self.path_entry.get())
        if self.change_path_cb is not None:
            ok, path = self.change_path_cb(self.path_entry.get())
            if not ok:
                self.set_path(path)

        self.path_entry.history_append(path)



"""
Preferences handling widgets
"""

from six import print_
from six.moves import tkinter as tk
from six.moves import tkinter_ttk as ttk

import tempfile
import copy
import collections

from . import config
from . import form
from . listmanager import ListManager
from . treewidget import TreeWidget
from . dialog import BrocoliDialog


class ConnectionConfigDialog(BrocoliDialog):
    """
    A dialog to configure connections
    """
    def __init__(self, master, connection_name='new-connection',
                 catalog_type=config.catalog_types[0],
                 root_path=tempfile.gettempdir(), isdefault=False,
                 catalog_config=None, **kwargs):
        self.connection_name = connection_name
        self.catalog_type = catalog_type
        self.root_path = root_path
        self.isdefault = isdefault
        self.catalog_config = catalog_config

        super(ConnectionConfigDialog, self).__init__(master, **kwargs)

    def body(self, master):
        a = 'e'
        tk.Label(master, text='Connection name:', anchor=a).grid(row=0,
                                                                 sticky='ew')
        tk.Label(master, text='Catalog type:', anchor=a).grid(row=1,
                                                              sticky='ew')
        tk.Label(master, text='Root path:', anchor=a).grid(row=2, sticky='ew')
        tk.Label(master, text='Default connection:', anchor=a).grid(row=3,
                                                                    sticky='ew')

        self.name = ttk.Entry(master)
        self.name.grid(row=0, column=1, sticky='ew')
        self.name.insert(0, self.connection_name)
        self.name.focus_set()

        self.catalog_cbox = ttk.Combobox(master, state='readonly',
                                         values=config.catalog_types + [])

        self.catalog_cbox.grid(row=1, column=1, sticky='ew')
        self.catalog_cbox.set(self.catalog_type)

        self.root_path_entry = tk.Entry(master)
        self.root_path_entry.grid(row=2, column=1, sticky='ew')
        self.root_path_entry.insert(0, self.root_path)

        self.isdefault_var = tk.IntVar()
        self.isdefault_var.set(self.isdefault)
        self.set_default = tk.Checkbutton(master, variable=self.isdefault_var,
                                          anchor='w')
        self.set_default.grid(row=3, column=1, sticky='ew')

        self.catalog_config_frame = tk.Frame(master)
        self.catalog_config_frame.grid(row=4, column=1, sticky='nsew')

        self.catalog_type_changed(catalog_config=self.catalog_config)
        self.catalog_cbox.bind('<<ComboboxSelected>>',
                               self.catalog_type_changed)

        self.result = None

        return self.name

    def apply(self):
        self.result = collections.OrderedDict([
            ('name', self.name.get()),
            ('catalog_type', self.catalog_cbox.get()),
            ('root_path', self.root_path_entry.get()),
            ('set_default', self.isdefault_var.get() != 0),
        ])

        config = collections.OrderedDict()
        for k, ff in self.config_fields.items():
            config[k] = ff.to_string()

        self.result.update(config)

    def to_config_dict(self):
        cfg = collections.OrderedDict()
        if self.result is None:
            return cfg

        new_conn = self.result
        name = new_conn['name']
        del new_conn['name']

        if new_conn['set_default']:
            cfg[config.SETTINGS] = {config.DEFAULT_CONNECTION: name}
        del new_conn['set_default']

        cfg['connection:' + name] = new_conn

        return cfg

    def catalog_type_changed(self, event=None, catalog_config=None):
        catalog_type = self.catalog_cbox.get()
        self.config_fields = config.catalog_dict[catalog_type].config_fields()

        if catalog_config is not None:
            for k, ff in self.config_fields.items():
                if k in catalog_config:
                    ff.from_string(catalog_config[k])

        master = self.catalog_config_frame.master
        self.catalog_config_frame.grid_remove()
        self.catalog_config_frame = form.FormFrame(master)
        self.catalog_config_frame.grid_fields(self.config_fields.values(),
                                              False)
        self.catalog_config_frame.columnconfigure(1, weight=1)
        self.catalog_config_frame.grid(row=4, columnspan=2, sticky='nsew')


class ConnectionManager(tk.Frame):
    """
    Displays connections and allows to add, remove and edit them
    """
    def __init__(self, master, cfg):
        tk.Frame.__init__(self, master)

        self.master = master
        self.cfg = cfg

        columns = ('catalog type', 'path', 'default connection')
        self.tree = ttk.Treeview(self, columns=columns,  selectmode='browse')

        ysb = ttk.Scrollbar(self, orient='vertical', command=self.tree.yview)
        xsb = ttk.Scrollbar(self, orient='horizontal', command=self.tree.xview)
        self.tree.configure(yscroll=ysb.set, xscroll=xsb.set)

        self.tree.heading('#0', text='name', anchor='w')
        self.tree.heading('catalog type', text='catalog type', anchor='w')
        self.tree.heading('path', text='path', anchor='w')
        self.tree.heading('default connection', text='default connection',
                          anchor='w')

        self.tree.grid(row=0, column=0, sticky='nsew')
        ysb.grid(row=0, column=1, sticky='ns')
        xsb.grid(row=1, column=0, sticky='ew')

        butbox = tk.Frame(self)
        butbox.grid(row=0, column=2, sticky='ns')
        self.newbut = tk.Button(butbox, text='Add', command=self.add)
        self.newbut.grid(row=0, column=0, sticky='ew')
        self.removebut = tk.Button(butbox, text='Remove', command=self.remove,
                                   state=tk.DISABLED)
        self.removebut.grid(row=1, column=0, sticky='ew')
        self.editbut = tk.Button(butbox, text='Edit', command=self.edit,
                                 state=tk.DISABLED)
        self.editbut.grid(row=2, column=0, sticky='ew')
        self.dupbut = tk.Button(butbox, text='Duplicate',
                                command=self.duplicate, state=tk.DISABLED)
        self.dupbut.grid(row=3, column=0, sticky='ew')

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        self.grid(sticky='nsew')

        self.tree.bind('<<TreeviewSelect>>', self.selchanged)

        self.insert_connections()

    def insert_connections(self):
        default = self.cfg[config.SETTINGS].get(config.DEFAULT_CONNECTION, None)

        for child in self.tree.get_children():
            self.tree.delete(child)

        for section in [s for s in self.cfg if s.startswith('connection:')]:
            conn = self.cfg[section]
            name = section.rsplit(':', 1)[1]
            isdefault = ''
            if name == default:
                isdefault = 'true'
            values = [conn['catalog_type'], conn['root_path'], isdefault]
            root_node = self.tree.insert('', 'end', iid=section, text=name,
                                         open=True, values=values)

    def add(self):
        n = ConnectionConfigDialog(self)
        new = n.result
        if new is None:
            return

        name = new['name']
        del new['name']

        if new['set_default']:
            self.cfg[config.SETTINGS][config.DEFAULT_CONNECTION] = name
        del new['set_default']

        self.cfg['connection:' + name] = new

        self.insert_connections()

    def remove(self):
        selected = self.tree.selection()[0]

        del self.cfg[selected]

        if config.DEFAULT_CONNECTION in self.cfg[config.SETTINGS] and \
           selected == ('connection:' +
                        self.cfg[config.SETTINGS][config.DEFAULT_CONNECTION]):
            del self.cfg[config.SETTINGS][config.DEFAULT_CONNECTION]

        self.tree.selection_set('')

        self.insert_connections()

    def edit(self, duplicate=False):
        selected = self.tree.selection()[0]
        item = self.tree.item(selected)

        name = item['text']
        cname = self.cfg['connection:' + name]
        if duplicate:
            name += ' (copy)'

        catalog_type, root_path, isdefault = item['values']

        if isdefault == '':
            isdefault = 0
        else:
            isdefault = 1

        n = ConnectionConfigDialog(self, name, catalog_type, root_path,
                                   isdefault, catalog_config=cname)
        new = n.result
        if new is None:
            return

        new_name = new['name']
        del new['name']
        if not duplicate and new_name != name:
            del self.cfg['connection:' + name]

        if new['set_default']:
            self.cfg[config.SETTINGS][config.DEFAULT_CONNECTION] = new_name
        elif self.cfg[config.SETTINGS].get(config.DEFAULT_CONNECTION, None) == \
                name:
            del self.cfg[config.SETTINGS][config.DEFAULT_CONNECTION]

        del new['set_default']

        self.cfg['connection:' + new_name] = new

        self.tree.selection_set('')
        self.insert_connections()

    def duplicate(self):
        self.edit(duplicate=True)

    def selchanged(self, event):
        buts = (self.editbut, self.removebut, self.dupbut)
        if self.tree.selection():
            for b in buts:
                b.config(state=tk.NORMAL)
        else:
            for b in buts:
                b.config(state=tk.DISABLED)


class ColumnManager(tk.Frame):
    def __init__(self, master, cfg, columns_def):
        tk.Frame.__init__(self, master)

        self.cfg = cfg
        self.columns_def = columns_def

        if 'display_columns' not in self.cfg[config.SETTINGS]:
            dcols = [k for k in self.columns_def.keys() if k != '#0']
            self.cfg[config.SETTINGS]['display_columns'] = ','.join(dcols)

        displayed = self.cfg[config.SETTINGS]['display_columns'].split(',')

        self.columns = {}

        j = 0
        for k, cd in self.columns_def.items():
            if k == "#0":
                continue

            v = tk.BooleanVar()
            v.set(k in displayed)
            b = tk.Checkbutton(self, text=cd.name, variable=v,
                               command=self.button_pressed)
            self.columns[k] = v
            b.grid(row=0, column=j)

            j += 1

    def button_pressed(self):
        # build a list of displayed columns in the columns_def order
        displayed = [k for k in self.columns_def if k != '#0' and
                     self.columns[k].get()]

        self.cfg[config.SETTINGS]['display_columns'] = ','.join(displayed)


class Preferences(BrocoliDialog):
    """
    Preferences dialog
    """
    def body(self, master):
        self.cfg = config.load_config()
        self.old_cfg = copy.deepcopy(self.cfg)

        self.notebook = ttk.Notebook(master)
        self.notebook.pack(expand=True, fill=tk.BOTH, side=tk.TOP)

        self.notebook.rowconfigure(0, weight=1)
        self.notebook.columnconfigure(0, weight=1)

        self.connection_manager = ConnectionManager(self.notebook, self.cfg)

        self.notebook.add(self.connection_manager, text='Connections')

        self.column_manager = ColumnManager(self.notebook, self.cfg,
                                            TreeWidget.columns_def)
        self.notebook.add(self.column_manager, text='Display')

        self.changed = False

        return self.notebook

    def apply(self):
        if self.cfg != self.old_cfg:
            config.save_config(self.cfg, update=False)
            self.changed = True



"""
Progression dialog classes
"""

from six.moves import tkinter as tk
from six.moves import tkinter_ttk as ttk
from six import print_

class UnboundedProgressDialog:
    """
    Displays a progression dialog without information on the completion of the
    task but with simple feeling of something evolving
    """
    def __init__(self, parent, opname, **kwargs):
        self.toplevel = tk.Toplevel(parent, **kwargs)
        self.toplevel.title(opname)
        self.toplevel.transient(parent)

        self.opname = opname

        self.label = tk.Label(self.toplevel, text=opname)
        self.label.pack()

        self.progress = ttk.Progressbar(self.toplevel, orient='horizontal',
                                        mode='indeterminate')

        self.progress.pack(expand=True, fill=tk.BOTH, side=tk.TOP)

    def step(self, speed=1):
        self.progress.step(speed)
        self.toplevel.update()

    def finish(self):
        self.toplevel.destroy()


class ProgressDialog:
    """
    Displays a dialog with a completion percentage
    """
    def __init__(self, parent, opname, interrupt=True, **kwargs):
        self.maximum = 100
        self.toplevel = tk.Toplevel(parent, **kwargs)
        self.toplevel.title(opname)
        self.toplevel.transient(parent)

        self.opname = opname

        self.count = tk.IntVar()

        self.label = tk.Label(self.toplevel, text=opname + ' progress: 0%')
        self.label.pack()

        self.count.set(-1)

        self.progress = ttk.Progressbar(self.toplevel, orient='horizontal',
                                        mode='determinate',
                                        variable=self.count,
                                        maximum=self.maximum)

        self.progress.pack(expand=True, fill=tk.BOTH, side=tk.TOP)
        self.interrupted = False
        if interrupt:
            interrupt_btn = tk.Button(self.toplevel, text='Interrupt',
                                      command=self.interrupt)
            interrupt_btn.pack()

    def interrupt(self):
        self.interrupted = True

    def set_message(self, message=None, percent=0):
        if message is not None:
            self.opname = message
        self.label.config(text=self.opname + ' progress: {}%'.format(percent))

    def set(self, value, maximum=100):
        value = int((100 * value) / maximum)
        if value == self.count.get():
            self.toplevel.update()
            return

        self.count.set(value)
        self.set_message(percent=value)
        self.toplevel.update()

    def finish(self):
        self.toplevel.destroy()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.finish()


def progress_from_generator(master, message, generator):
    """
    Builds a ProgressDialog evolving from a generator that yields pairs of
    values in the form (current, total)
    """
    with ProgressDialog(master, message) as progress:
        for p, n in generator:
            progress.set(p, n)

            if progress.interrupted:
                return


def unbounded_progress_from_generator(master, message, generator):
    """
    Builds a ProgressDialog evolving from a generator yields (no matter what
    the generator yields)
    """
    with UnboundedProgressDialog(master, message) as progress:
        for _ in generator:
            progress.step(4)
